# This file contains python code for the Google Earth Engine datasets used in the Whisp pacakge.

# If you are running a bespoke analysis including your own datasets see also the main README.md file.

# Key aspects to include in the code for each function are:
# a) a suffix of ' _prep' and
# b) a prefix of "nXX_" if it is national/sub-national dataset (where XX is replaced by that country code), or a prefix of 'g_' if it covers more than one country.
# c) a name for your image, defined by ".rename('add_your_image_name_here')". This becomes the column header in the output table.

# Tips:
#  -Avoid getInfo() and for loops to speed up processing by keeping everything in the Earth Engine API.
#  -For all the above you will need to be running the package in editable mode for these local changes to take effect.
#   Editable mode runs the package locally and thus changes to any files are reflected immediately.

import ee

# ee.Authenticate()
# ee.Initialize()

from datetime import datetime

# from openforis_whisp.parameters.config_runtime import (
#     geometry_area_column,
# )  # ideally make relative import statement

# defining here instead of importing from config_runtime, to allow functioning as more of a standalone script
geometry_area_column = "Area"

# Calculate current year once at module load time (not in functions)
# This avoids repeated datetime calls and potential .getInfo() calls
CURRENT_YEAR = datetime.now().year
CURRENT_YEAR_2DIGIT = CURRENT_YEAR % 100  # Last two digits for RADD datasets

import inspect

import logging

# Configure logging
logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)


def get_logger(name):
    return logging.getLogger(name)


# Add datasets below

# tree cover datasets


# ESA_TC_2020
def g_esa_worldcover_trees_prep():
    esa_worldcover_2020_raw = ee.Image("ESA/WorldCover/v100/2020")
    esa_worldcover_trees_2020 = esa_worldcover_2020_raw.eq(95).Or(
        esa_worldcover_2020_raw.eq(10)
    )  # get trees and mnangroves
    return esa_worldcover_trees_2020.rename("ESA_TC_2020").selfMask()


# EUFO_2020
def g_jrc_gfc_2020_prep():
    # JRC GFC2020 V3 is a single Image with band 'Map'
    jrc_gfc2020 = ee.Image("JRC/GFC2020/V3").select("Map")
    return jrc_gfc2020.rename("EUFO_2020").selfMask()


# GFC_TC_2020
def g_glad_gfc_10pc_prep():
    gfc = ee.Image("UMD/hansen/global_forest_change_2024_v1_12")
    gfc_treecover2000 = gfc.select(["treecover2000"])
    gfc_loss2001_2020 = gfc.select(["lossyear"]).lte(20)
    gfc_treecover2020 = gfc_treecover2000.where(gfc_loss2001_2020.eq(1), 0)
    return gfc_treecover2020.gt(10).rename("GFC_TC_2020").selfMask()


# GLAD_Primary
def g_glad_pht_prep():
    primary_ht_forests2001_raw = ee.ImageCollection(
        "UMD/GLAD/PRIMARY_HUMID_TROPICAL_FORESTS/v1"
    )
    primary_ht_forests2001 = (
        primary_ht_forests2001_raw.select("Primary_HT_forests").mosaic().selfMask()
    )
    gfc = ee.Image("UMD/hansen/global_forest_change_2024_v1_12")
    gfc_loss2001_2020 = gfc.select(["lossyear"]).lte(20)
    return (
        primary_ht_forests2001.where(gfc_loss2001_2020.eq(1), 0)
        .rename("GLAD_Primary")
        .selfMask()
    )


# TMF_undist (undistrubed forest in 2020)
def g_jrc_tmf_undisturbed_prep():
    TMF_undist_2020 = (
        ee.ImageCollection("projects/JRC/TMF/v1_2024/AnnualChanges")
        .select("Dec2020")
        .mosaic()
        .eq(1)
    )  # update from https://github.com/forestdatapartnership/whisp/issues/42
    return TMF_undist_2020.rename("TMF_undist").selfMask()


# Forest Persistence FDaP
def g_fdap_forest_prep():
    fdap_forest_raw = ee.Image(
        "projects/forestdatapartnership/assets/community_forests/ForestPersistence_2020"
    )
    fdap_forest = fdap_forest_raw.gt(0.75)
    return fdap_forest.rename("Forest_FDaP").selfMask()


#########################primary forest
# EUFO JRC Global forest type - primary
def g_gft_primary_prep():
    gft_raw = ee.ImageCollection("JRC/GFC2020_subtypes/V0").mosaic()
    gft_primary = gft_raw.eq(10)
    return gft_primary.rename("GFT_primary").selfMask()


# Intact Forest Landscape 2020
def g_ifl_2020_prep():
    IFL_2020 = ee.Image("users/potapovpeter/IFL_2020")
    return IFL_2020.rename("IFL_2020").selfMask()


# European Primary Forest Dataset
def g_epfd_prep():
    EPFD = ee.FeatureCollection("HU_BERLIN/EPFD/V2/polygons")
    EPFD_binary = ee.Image().paint(EPFD, 1)
    return EPFD_binary.rename("European_Primary_Forest").selfMask()


# EUFO JRC Global forest type - naturally regenerating planted/plantation forests
def g_gft_nat_reg_prep():
    gft_raw = ee.ImageCollection("JRC/GFC2020_subtypes/V0").mosaic()
    gft_nat_reg = gft_raw.eq(1)
    return gft_nat_reg.rename("GFT_naturally_regenerating").selfMask()


#########################planted and plantation forests

# EUFO JRC Global forest type - planted/plantation forests
def g_gft_plantation_prep():
    gft_raw = ee.ImageCollection("JRC/GFC2020_subtypes/V0").mosaic()
    gft_plantation = gft_raw.eq(20)
    return gft_plantation.rename("GFT_planted_plantation").selfMask()


def g_iiasa_planted_prep():
    iiasa = ee.Image("projects/sat-io/open-datasets/GFM/FML_v3-2")
    iiasa_PL = iiasa.eq(31).Or(iiasa.eq(32))
    return iiasa_PL.rename("IIASA_planted_plantation").selfMask()


#########################TMF regrowth in 2023
def g_tmf_regrowth_prep():
    # Load the TMF Degradation annual product
    TMF_AC = ee.ImageCollection("projects/JRC/TMF/v1_2024/AnnualChanges").mosaic()
    TMF_AC_2023 = TMF_AC.select("Dec2023")
    Regrowth_TMF = TMF_AC_2023.eq(4)
    return Regrowth_TMF.rename("TMF_regrowth_2023").selfMask()


############tree crops

# TMF_plant (plantations in 2020)
def g_jrc_tmf_plantation_prep():
    transition = ee.ImageCollection(
        "projects/JRC/TMF/v1_2024/TransitionMap_Subtypes"
    ).mosaic()
    deforestation_year = ee.ImageCollection(
        "projects/JRC/TMF/v1_2024/DeforestationYear"
    ).mosaic()
    plantation = (transition.gte(81)).And(transition.lte(86))
    plantation_2020 = plantation.where(
        deforestation_year.gte(2021), 0
    )  # update from https://github.com/forestdatapartnership/whisp/issues/42
    return plantation_2020.rename("TMF_plant").selfMask()


# # Oil_palm_Descals
# NB updated to Descals et al 2024 paper (as opposed to Descals et al 2021 paper)
def g_creaf_descals_palm_prep():
    # Load the Global Oil Palm Year of Plantation image and mosaic it
    img = (
        ee.ImageCollection(
            "projects/ee-globaloilpalm/assets/shared/GlobalOilPalm_YoP_2021"
        )
        .mosaic()
        .select("minNBR_date")
    ).selfMask()

    # Calculate the year of plantation and select all below and including 2020
    oil_palm_plantation_year = img.divide(365).add(1970).floor().lte(2020)

    # Create a mask for plantations in the year 2020 or earlier
    plantation_2020 = oil_palm_plantation_year.lte(2020)
    return plantation_2020.rename("Oil_palm_Descals").selfMask()


# Cocoa_ETH
def g_eth_kalischek_cocoa_prep():
    return (
        ee.Image("projects/ee-nk-cocoa/assets/cocoa_map_threshold_065")
        .rename("Cocoa_ETH")
        .selfMask()
    )


# fdap datasets

# Thresholds and model info here https://github.com/google/forest-data-partnership/blob/main/models/README.md

# Oil Palm FDaP
def g_fdap_palm_prep():
    fdap_palm2020_model_raw = ee.ImageCollection(
        "projects/forestdatapartnership/assets/palm/model_2025a"
    )
    fdap_palm = (
        fdap_palm2020_model_raw.filterDate("2020-01-01", "2020-12-31")
        .mosaic()
        .gt(0.88)  # Precision and recall ~78% at 0.88 threshold.
    )
    return fdap_palm.rename("Oil_palm_FDaP").selfMask()


def g_fdap_palm_2023_prep():
    fdap_palm2020_model_raw = ee.ImageCollection(
        "projects/forestdatapartnership/assets/palm/model_2025a"
    )
    fdap_palm = (
        fdap_palm2020_model_raw.filterDate("2023-01-01", "2023-12-31")
        .mosaic()
        .gt(0.88)  # Precision and recall ~78% at 0.88 threshold.
    )
    return fdap_palm.rename("Oil_palm_2023_FDaP").selfMask()


# Cocoa FDaP
def g_fdap_cocoa_prep():
    fdap_cocoa2020_model_raw = ee.ImageCollection(
        "projects/forestdatapartnership/assets/cocoa/model_2025a"
    )
    fdap_cocoa = (
        fdap_cocoa2020_model_raw.filterDate("2020-01-01", "2020-12-31")
        .mosaic()
        .gt(0.96)  # Precision and recall ~87% 0.96 threshold.
    )
    return fdap_cocoa.rename("Cocoa_FDaP").selfMask()


def g_fdap_cocoa_2023_prep():
    fdap_cocoa2020_model_raw = ee.ImageCollection(
        "projects/forestdatapartnership/assets/cocoa/model_2025a"
    )
    fdap_cocoa = (
        fdap_cocoa2020_model_raw.filterDate("2023-01-01", "2023-12-31")
        .mosaic()
        .gt(0.96)  # Precision and recall ~87% 0.96 threshold.
    )
    return fdap_cocoa.rename("Cocoa_2023_FDaP").selfMask()


# Rubber FDaP
def g_fdap_rubber_prep():
    fdap_rubber2020_model_raw = ee.ImageCollection(
        "projects/forestdatapartnership/assets/rubber/model_2025a"
    )
    fdap_rubber = (
        fdap_rubber2020_model_raw.filterDate("2020-01-01", "2020-12-31")
        .mosaic()
        .gt(0.59)  # Precision and recall ~80% 0.59 threshold.
    )
    return fdap_rubber.rename("Rubber_FDaP").selfMask()


def g_fdap_rubber_2023_prep():
    fdap_rubber2020_model_raw = ee.ImageCollection(
        "projects/forestdatapartnership/assets/rubber/model_2025a"
    )
    fdap_rubber = (
        fdap_rubber2020_model_raw.filterDate("2023-01-01", "2023-12-31")
        .mosaic()
        .gt(0.59)  # Threshold for Rubber
    )
    return fdap_rubber.rename("Rubber_2023_FDaP").selfMask()


# # Coffee FDaP
def g_fdap_coffee_2020_prep():
    # Load the coffee model for 2020
    collection = ee.ImageCollection(
        "projects/forestdatapartnership/assets/coffee/model_2025a"
    )

    # Filter the collection for the year 2020 and create a binary mask
    coffee_2020 = (
        collection.filterDate("2020-01-01", "2020-12-31")
        .mosaic()
        .gt(0.99)  # Precision and recall ~54% 0.99 threshold.
    )

    return coffee_2020.rename("Coffee_FDaP").selfMask()


def g_fdap_coffee_2023_prep():
    # Load the coffee model for 2020
    collection = ee.ImageCollection(
        "projects/forestdatapartnership/assets/coffee/model_2025a"
    )

    # Filter the collection for the year 2023 and create a binary mask
    coffee_2023 = (
        collection.filterDate("2023-01-01", "2023-12-31")
        .mosaic()
        .gt(0.99)  # Precision and recall ~54% 0.99 threshold.
    )
    return coffee_2023.rename("Coffee_FDaP_2023").selfMask()


# Rubber_RBGE  - from Royal Botanical Gardens of Edinburgh (RBGE) NB for 2021
def g_rbge_rubber_prep():
    return (
        ee.Image(
            "users/wangyxtina/MapRubberPaper/rRubber10m202122_perc1585DifESAdist5pxPF"
        )
        .rename("Rubber_RBGE")
        .selfMask()
    )


# soy 2020 South America
def g_soy_song_2020_prep():
    return (
        ee.Image("projects/glad/soy_annual_SA/2020").rename("Soy_Song_2020").selfMask()
    )


##############
# ESRI 2023

# ESRI 2023 - Tree Cover
def g_esri_2023_tc_prep():
    esri_lulc10_raw = ee.ImageCollection(
        "projects/sat-io/open-datasets/landcover/ESRI_Global-LULC_10m_TS"
    )
    esri_lulc10_TC = (
        esri_lulc10_raw.filterDate("2023-01-01", "2023-12-31").mosaic().eq(2)
    )
    return esri_lulc10_TC.rename("ESRI_2023_TC").selfMask()


# ESRI 2023 - Crop
def g_esri_2020_2023_crop_prep():
    esri_lulc10_raw = ee.ImageCollection(
        "projects/sat-io/open-datasets/landcover/ESRI_Global-LULC_10m_TS"
    )
    esri_lulc10_crop_2020 = (
        esri_lulc10_raw.filterDate("2020-01-01", "2020-12-31").mosaic().eq(5)
    )
    esri_lulc10_crop_2023 = (
        esri_lulc10_raw.filterDate("2023-01-01", "2023-12-31").mosaic().eq(5)
    )

    newCrop = esri_lulc10_crop_2023.And(esri_lulc10_crop_2020.Not())

    return newCrop.rename("ESRI_crop_gain_2020_2023").selfMask()


#### disturbances by year

# RADD_year_2019 to RADD_year_< current year >
def g_radd_year_prep():
    radd = ee.ImageCollection("projects/radar-wur/raddalert/v1")
    radd_date = (
        radd.filterMetadata("layer", "contains", "alert").select("Date").mosaic()
    )
    start_year = 19
    current_year = CURRENT_YEAR_2DIGIT

    def make_band(year, img_stack):
        start = year * 1000
        end = year * 1000 + 365
        radd_year = (
            radd_date.updateMask(radd_date.gte(start))
            .updateMask(radd_date.lte(end))
            .gt(0)
            .rename("RADD_year_" + "20" + str(year))
            .selfMask()
        )
        return ee.Image(img_stack).addBands(radd_year)

    years = ee.List.sequence(start_year, current_year)
    first_year = ee.Number(years.get(0))
    start = first_year.multiply(1000)
    end = first_year.multiply(1000).add(365)
    band_name = ee.String("RADD_year_").cat("20").cat(first_year.format("%02d"))
    first_band = (
        radd_date.updateMask(radd_date.gte(start))
        .updateMask(radd_date.lte(end))
        .gt(0)
        .rename(band_name)
        .selfMask()
    )

    def make_band(year, img_stack):
        year_num = ee.Number(year)
        start = year_num.multiply(1000)
        end = year_num.multiply(1000).add(365)
        band_name = ee.String("RADD_year_").cat("20").cat(year_num.format("%02d"))
        radd_year = (
            radd_date.updateMask(radd_date.gte(start))
            .updateMask(radd_date.lte(end))
            .gt(0)
            .rename(band_name)
            .selfMask()
        )
        return ee.Image(img_stack).addBands(radd_year)

    img_stack = years.slice(1).iterate(make_band, first_band)
    return ee.Image(img_stack)


# TMF_def_2000 to TMF_def_2023
def g_tmf_def_per_year_prep():
    # Load the TMF Deforestation annual product
    tmf_def = ee.ImageCollection("projects/JRC/TMF/v1_2024/DeforestationYear").mosaic()
    img_stack = None
    # Generate an image based on GFC with one band of forest tree loss per year from 2001 to 2022
    for i in range(0, 24 + 1):
        year_num = ee.Number(2000 + i)
        band_name = ee.String("TMF_def_").cat(year_num.format("%d"))
        tmf_def_year = tmf_def.eq(year_num).rename(band_name).selfMask()
        if img_stack is None:
            img_stack = tmf_def_year
        else:
            img_stack = img_stack.addBands(tmf_def_year)
    return img_stack


# TMF_deg_2000 to TMF_deg_2023
def g_tmf_deg_per_year_prep():
    # Load the TMF Degradation annual product
    tmf_def = ee.ImageCollection("projects/JRC/TMF/v1_2024/DegradationYear").mosaic()
    img_stack = None
    # Generate an image based on GFC with one band of forest tree loss per year from 2001 to 2022
    for i in range(0, 24 + 1):
        year_num = ee.Number(2000 + i)
        band_name = ee.String("TMF_deg_").cat(year_num.format("%d"))
        tmf_def_year = tmf_def.eq(year_num).rename(band_name).selfMask()
        if img_stack is None:
            img_stack = tmf_def_year
        else:
            img_stack = img_stack.addBands(tmf_def_year)
    return img_stack


# GFC_loss_year_2001 to GFC_loss_year_2023 (correct for version 11)
def g_glad_gfc_loss_per_year_prep():
    # Load the Global Forest Change dataset
    gfc = ee.Image("UMD/hansen/global_forest_change_2024_v1_12")
    img_stack = None
    # Generate an image based on GFC with one band of forest tree loss per year from 2001 to 2022
    for i in range(1, 24 + 1):
        year_num = ee.Number(2000 + i)
        band_name = ee.String("GFC_loss_year_").cat(year_num.format("%d"))
        gfc_loss_year = (
            gfc.select(["lossyear"]).eq(i).And(gfc.select(["treecover2000"]).gt(10))
        )
        gfc_loss_year = gfc_loss_year.rename(band_name).selfMask()
        if img_stack is None:
            img_stack = gfc_loss_year
        else:
            img_stack = img_stack.addBands(gfc_loss_year)
    return img_stack


# MODIS_fire_2000 to MODIS_fire_< current year >
def g_modis_fire_prep():
    modis_fire = ee.ImageCollection("MODIS/061/MCD64A1")
    start_year = 2000

    # Determine the last available year by checking the latest image in the collection
    last_image = modis_fire.sort("system:time_start", False).first()
    last_date = ee.Date(last_image.get("system:time_start"))
    end_year = last_date.get("year").getInfo()

    img_stack = None

    for year in range(start_year, end_year + 1):
        year_num = ee.Number(year)
        band_name = ee.String("MODIS_fire_").cat(year_num.format("%d"))
        date_st = f"{year}-01-01"
        date_ed = f"{year}-12-31"
        modis_year = (
            modis_fire.filterDate(date_st, date_ed)
            .mosaic()
            .select(["BurnDate"])
            .gte(0)
            .rename(band_name)
            .selfMask()
        )
        img_stack = modis_year if img_stack is None else img_stack.addBands(modis_year)

    return img_stack


# ESA_fire_2000 to ESA_fire_2020
def g_esa_fire_prep():
    esa_fire = ee.ImageCollection("ESA/CCI/FireCCI/5_1")
    start_year = 2001

    # Determine the last available year by checking the latest image in the collection
    last_image = esa_fire.sort("system:time_start", False).first()
    last_date = ee.Date(last_image.get("system:time_start"))
    end_year = last_date.get("year").getInfo()

    img_stack = None

    for year in range(start_year, end_year + 1):
        year_num = ee.Number(year)
        band_name = ee.String("ESA_fire_").cat(year_num.format("%d"))
        date_st = f"{year}-01-01"
        date_ed = f"{year}-12-31"
        esa_year = (
            esa_fire.filterDate(date_st, date_ed)
            .mosaic()
            .select(["BurnDate"])
            .gte(0)
            .rename(band_name)
            .selfMask()
        )
        img_stack = esa_year if img_stack is None else img_stack.addBands(esa_year)

    return img_stack


#### disturbances combined (split into before and after 2020)

# RADD_after_2020
def g_radd_after_2020_prep():
    radd = ee.ImageCollection("projects/radar-wur/raddalert/v1")

    radd_date = (
        radd.filterMetadata("layer", "contains", "alert").select("Date").mosaic()
    )
    # date of avaialbility
    start_year = 21  ## (starts 2019 in Africa, then 2020 for S America and Asia: https://data.globalforestwatch.org/datasets/gfw::deforestation-alerts-radd/about)

    # Use pre-calculated current year (avoids repeated datetime calls)
    current_year = CURRENT_YEAR_2DIGIT
    start = start_year * 1000
    end = current_year * 1000 + 365
    return (
        radd_date.updateMask(radd_date.gte(start))
        .updateMask(radd_date.lte(end))
        .gt(0)
        .rename("RADD_after_2020")
    ).selfMask()


# RADD_before_2020
def g_radd_before_2020_prep():
    radd = ee.ImageCollection("projects/radar-wur/raddalert/v1")

    radd_date = (
        radd.filterMetadata("layer", "contains", "alert").select("Date").mosaic()
    )
    # date of avaialbility
    start_year = 19  ## (starts 2019 in Africa, then 2020 for S America and Asia: https://data.globalforestwatch.org/datasets/gfw::deforestation-alerts-radd/about)

    start = start_year * 1000
    end = 20 * 1000 + 365
    return (
        radd_date.updateMask(radd_date.gte(start))
        .updateMask(radd_date.lte(end))
        .gt(0)
        .rename("RADD_before_2020")
    ).selfMask()


# # DIST_after_2020
# # alerts only for after 2020 currently so need to use date
# def glad_dist_after_2020_prep():

#     # Load the vegetation disturbance collections
#     VEGDISTSTATUS = ee.ImageCollection(
#         "projects/glad/HLSDIST/current/VEG-DIST-STATUS"
#     ).mosaic()

#     # Key for high-confidence alerts (values 3, 6, 7, 8)
#     high_conf_values = [3, 6, 7, 8]

#     # Create high-confidence mask
#     dist_high_conf = VEGDISTSTATUS.remap(
#         high_conf_values, [1] * len(high_conf_values), 0
#     )

#     return dist_high_conf.updateMask(jrc_gfc_2020_prep()).rename(
#         "DIST_after_2020"
#     )  # Mask alerts to forest and rename band


# TMF_deg_before_2020
def g_tmf_deg_before_2020_prep():
    tmf_deg = ee.ImageCollection("projects/JRC/TMF/v1_2024/DegradationYear").mosaic()
    return (
        (tmf_deg.lte(2020))
        .And(tmf_deg.gte(2000))
        .rename("TMF_deg_before_2020")
        .selfMask()
    )


# TMF_deg_after_2020
def g_tmf_deg_after_2020_prep():
    tmf_deg = ee.ImageCollection("projects/JRC/TMF/v1_2024/DegradationYear").mosaic()
    return tmf_deg.gt(2020).rename("TMF_deg_after_2020").selfMask()


# tmf_def_before_2020
def g_tmf_def_before_2020_prep():
    tmf_def = ee.ImageCollection("projects/JRC/TMF/v1_2024/DeforestationYear").mosaic()
    return (
        (tmf_def.lte(2020))
        .And(tmf_def.gte(2000))
        .rename("TMF_def_before_2020")
        .selfMask()
    )


# tmf_def_after_2020
def g_tmf_def_after_2020_prep():
    tmf_def = ee.ImageCollection("projects/JRC/TMF/v1_2024/DeforestationYear").mosaic()
    return tmf_def.gt(2020).rename("TMF_def_after_2020").selfMask()


# GFC_loss_before_2020 (loss within 10 percent cover; includes 2020; correct for version 11)
def g_glad_gfc_loss_before_2020_prep():
    # Load the Global Forest Change dataset
    gfc = ee.Image("UMD/hansen/global_forest_change_2024_v1_12")
    gfc_loss = (
        gfc.select(["lossyear"]).lte(20).And(gfc.select(["treecover2000"]).gt(10))
    )
    return gfc_loss.rename("GFC_loss_before_2020").selfMask()


# GFC_loss_after_2020 (loss within 10 percent cover; correct for version 11)
def g_glad_gfc_loss_after_2020_prep():
    # Load the Global Forest Change dataset
    gfc = ee.Image("UMD/hansen/global_forest_change_2024_v1_12")
    gfc_loss = gfc.select(["lossyear"]).gt(20).And(gfc.select(["treecover2000"]).gt(10))
    return gfc_loss.rename("GFC_loss_after_2020").selfMask()


# MODIS_fire_before_2020
def g_modis_fire_before_2020_prep():
    modis_fire = ee.ImageCollection("MODIS/061/MCD64A1")
    start_year = 2000
    end_year = 2020
    date_st = str(start_year) + "-01-01"
    date_ed = str(end_year) + "-12-31"
    return (
        modis_fire.filterDate(date_st, date_ed)
        .mosaic()
        .select(["BurnDate"])
        .gte(0)
        .rename("MODIS_fire_before_2020")
    ).selfMask()


# MODIS_fire_after_2020
def g_modis_fire_after_2020_prep():
    modis_fire = ee.ImageCollection("MODIS/061/MCD64A1")
    start_year = 2021
    # Use pre-calculated current year (avoids repeated datetime calls)
    end_year = CURRENT_YEAR - 1  # Use year - 1 to ensure data availability
    date_st = str(start_year) + "-01-01"
    date_ed = str(end_year) + "-12-31"
    return (
        modis_fire.filterDate(date_st, date_ed)
        .mosaic()
        .select(["BurnDate"])
        .gte(0)
        .rename("MODIS_fire_after_2020")
    ).selfMask()


# ESA_fire_before_2020
def g_esa_fire_before_2020_prep():
    esa_fire = ee.ImageCollection("ESA/CCI/FireCCI/5_1")
    start_year = 2000
    end_year = 2020
    date_st = str(start_year) + "-01-01"
    date_ed = str(end_year) + "-12-31"
    return (
        esa_fire.filterDate(date_st, date_ed)
        .mosaic()
        .select(["BurnDate"])
        .gte(0)
        .rename("ESA_fire_before_2020")
    ).selfMask()


#########################logging concessions
# http://data.globalforestwatch.org/datasets?q=logging&sort_by=relevance
def g_logging_concessions_before_2020_prep():
    RCA = ee.FeatureCollection(
        "projects/ee-whisp/assets/logging/RCA_Permis_dExploitation_et_dAmenagement"
    )
    RCA_binary = ee.Image().paint(RCA, 1)
    CMR = ee.FeatureCollection(
        "projects/ee-whisp/assets/logging/Cameroon_Forest_Management_Units"
    )
    CMR_binary = ee.Image().paint(CMR, 1)
    Eq_G = ee.FeatureCollection(
        "projects/ee-whisp/assets/logging/Equatorial_Guinea_logging_concessions"
    )
    Eq_G_binary = ee.Image().paint(Eq_G, 1)
    DRC = ee.FeatureCollection(
        "projects/ee-whisp/assets/logging/DRC_Forest_concession_agreements"
    )
    DRC_binary = ee.Image().paint(DRC, 1)
    Liberia = ee.FeatureCollection(
        "projects/ee-whisp/assets/logging/Liberia_Forest_Management_Contracts"
    )
    Liberia_binary = ee.Image().paint(Liberia, 1)
    RoC = ee.FeatureCollection(
        "projects/ee-whisp/assets/logging/Republic_of_the_Congo_logging_concessions"
    )
    Roc_binary = ee.Image().paint(RoC, 1)
    Sarawak = ee.FeatureCollection(
        "projects/ee-whisp/assets/logging/Sarawak_logging_concessions"
    )
    Sarawak_binary = ee.Image().paint(Sarawak, 1)
    logging_concessions_binary = ee.ImageCollection(
        [
            RCA_binary,
            CMR_binary,
            Eq_G_binary,
            DRC_binary,
            Liberia_binary,
            Roc_binary,
            Sarawak_binary,
        ]
    ).mosaic()

    return logging_concessions_binary.rename("GFW_logging_before_2020").selfMask()


#########################national datasets

# nBR Brazil

# ### nBR Natural forests in 2020:

# %%
# [Official NFMS dataset] INPE/EMBRAPA TerraClass land use/cover in the Amazon biome, 2020
# Subsetting criteria: primary forests (DN=1) and secondary forests (DN=2) // secondary forests are those recovering from deforestation
# the resulting dataset shows primary and secondary forest cover in 2020 (mostly by August 2020)

##########################primary forests###############################################
def nbr_terraclass_amz20_primary_prep():
    tcamz20 = ee.Image("projects/ee-whisp/assets/NBR/terraclass_amz_2020")
    tcamz20_f = tcamz20.eq(1)
    return tcamz20_f.rename("nBR_INPE_TC_primary_forest_Amazon_2020").selfMask()


# [Official NFMS dataset] Brazilian Forest Service dataset on natural forest cover from PRODES and TerraClass data, base year 2022
# Subsetting criteria: ano_desmat > 2020 and nom_class = 'Floresta'
# the resulting datasets show primary forest cover in 2020 for the Pantanal, Caatinga, Atlantic Forest and Pampa biomes.
# the resulting dataset shows primary and secondary forest cover in 2020 for the Cerrado biome (TerraClass 2020)
# For the Amazon, best to use Terraclass 2020 directly, because the BFS used TerraClass 2014.

# Pantanal
def nbr_bfs_ptn_f20_prep():
    bfs_fptn20 = ee.FeatureCollection("projects/ee-whisp/assets/NBR/bfs_ptn_2020")

    bfs_fptn20_binary = ee.Image().paint(bfs_fptn20, 1)
    return bfs_fptn20_binary.rename("nBR_BFS_primary_forest_Pantanal_2020").selfMask()


# Caatinga - filtered with QGIS because the original geodatabase is too large to export as a shapefile (GEE accepted format)
## couldn't convert it to asset, working on it (Error: Primary geometry of feature '306862' has 2454627 vertices, above the limit of 1000000 vertices. (Error code: 3)
def nbr_bfs_caat_f20_prep():
    bfs_fcaat20 = ee.FeatureCollection("projects/ee-whisp/assets/NBR/bfs_caat_2020")
    bfs_fcaat20_binary = ee.Image().paint(bfs_fcaat20, 1)
    return bfs_fcaat20_binary.rename("nBR_BFS_primary_forest_Caatinga_2020").selfMask()


# Atlantic Forest - filtered with QGIS because the original geodatabase is too large to export as a shapefile (GEE accepted format)
def nbr_bfs_atlf_f20_prep():
    bfs_fatlf20 = ee.FeatureCollection("projects/ee-whisp/assets/NBR/bfs_atlf_2020")
    bfs_fatlf20_binary = ee.Image().paint(bfs_fatlf20, 1)
    return bfs_fatlf20_binary.rename(
        "nBR_BFS_primary_forest_AtlanticForest_2020"
    ).selfMask()


# Pampa - filtered in QGIS to save some storage space
def nbr_bfs_pmp_f20_prep():
    bfs_fpmp20 = ee.FeatureCollection("projects/ee-whisp/assets/NBR/bfs_pmp_2020")
    bfs_fpmp20_binary = ee.Image().paint(bfs_fpmp20, 1)
    return bfs_fpmp20_binary.rename("nBR_BFS_primary_forest_Pampa_2020").selfMask()


##########################secondary forests###############################################
def nbr_terraclass_amz20_secondary_prep():
    tcamz20 = ee.Image("projects/ee-whisp/assets/NBR/terraclass_amz_2020")
    tcamz20_f = tcamz20.eq(2)
    return tcamz20_f.rename("nBR_INPE_TC_secondary_forest_Amazon_2020").selfMask()


# Cerrado - filtered with QGIS because the original geodatabase is too large to export as a shapefile (GEE accepted format)
def nbr_bfs_cer_f20_prep():
    bfs_fcer20 = ee.FeatureCollection("projects/ee-whisp/assets/NBR/bfs_cerr_2020")
    bfs_fcer20_binary = ee.Image().paint(bfs_fcer20, 1)
    return bfs_fcer20_binary.rename(
        "nBR_BFS_primary_and_secondary_forest_Cerrado_2020"
    ).selfMask()


# %%
# [non-official dataset by MapBiomas multisector initiative]
# land use/cover from 1985 up to 2023, collection 9
# Subsetting criteria: classification_2020 = Forest formation (DN=3), Savanna Formation (DN=4, forest according to BR definition), Mangrove (DN=5), Floodable Forest (DN=6), Wooded Sandbank veg (DN=49)
# the resulting dataset shows forest cover in 2020, without distinguishing between primary and secondary forests
def nbr_mapbiomasc9_f20_prep():
    mapbiomasc9_20 = ee.Image(
        "projects/mapbiomas-public/assets/brazil/lulc/collection9/mapbiomas_collection90_integration_v1"
    ).select("classification_2020")
    mapbiomasc9_20_forest = (
        mapbiomasc9_20.eq(3)
        .Or(mapbiomasc9_20.eq(4))
        .Or(mapbiomasc9_20.eq(5))
        .Or(mapbiomasc9_20.eq(6))
        .Or(mapbiomasc9_20.eq(49))
    )
    return mapbiomasc9_20_forest.rename(
        "nBR_MapBiomas_col9_forest_Brazil_2020"
    ).selfMask()


# ### ########################NBR plantation forest in 2020:#######################################

# [Official NFMS dataset] INPE/EMBRAPA TerraClass land use/cover in the Amazon biome, 2020
# Subsetting criteria: silviculture (DN=9)
# the resulting dataset shows monospecific commercial plantations, mostly eucalyptus and pinus.
def nbr_terraclass_amz20_silv_prep():
    tcamz20 = ee.Image("projects/ee-whisp/assets/NBR/terraclass_amz_2020")
    tcamz20_silviculture = tcamz20.eq(9)
    return tcamz20_silviculture.rename("nBR_INPE_TCsilviculture_Amazon_2020").selfMask()


# [Official NFMS dataset] INPE/EMBRAPA TerraClass land use/cover in the Cerrado biome, 2020
# Subsetting criteria: silviculture (DN=9)
# the resulting dataset shows monospecific commercial plantations, mostly eucalyptus and pinus.
def nbr_terraclass_silv_cer20_prep():
    tccer20 = ee.Image("projects/ee-whisp/assets/NBR/terraclass_cer_2020")
    tccer20_silviculture = tccer20.eq(9)
    return tccer20_silviculture.rename(
        "nBR_INPE_TCsilviculture_Cerrado_2020"
    ).selfMask()


# [non-official dataset by MapBiomas multisector initiative]
# land use/cover from 1985 up to 2023, collection 9
# Subsetting criteria: 'classification_2020' = Forest plantation (DN=9)
# the resulting dataset shows forest plantation in 2020
def nbr_mapbiomasc9_silv20_prep():
    mapbiomasc9_20 = ee.Image(
        "projects/mapbiomas-public/assets/brazil/lulc/collection9/mapbiomas_collection90_integration_v1"
    ).select("classification_2020")
    mapbiomasc9_20_silviculture = mapbiomasc9_20.eq(9)
    return mapbiomasc9_20_silviculture.rename(
        "nBR_MapBiomas_col9_silviculture_Brazil_2020"
    ).selfMask()


################ ### NBR Disturbances before 2020:########################################

# [Official NFMS dataset] INPE PRODES data up to 2023
# Subsetting criteria: DN = [0, 2, 4, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 50, 51, 52, 53, 54, 55, 56, 57, 58, 59, 60];

# the resulting dataset shows deforestation and conversion of OWL and OL up to 2020 (mostly August 2020), including residues (omission errors corrections)
def nbr_prodes_before_2020_prep():
    prodes = ee.Image("projects/ee-whisp/assets/NBR/prodes_brasil_2023")
    prodes_before_20_dn = [
        0,
        2,
        4,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        50,
        51,
        52,
        53,
        54,
        55,
        56,
        57,
        58,
        59,
        60,
    ]
    prodes_before_20_mask = prodes.remap(
        prodes_before_20_dn, [1] * len(prodes_before_20_dn)
    )  # .eq(1)
    return prodes_before_20_mask.rename(
        "nBR_PRODES_deforestation_Brazil_before_2020"
    ).selfMask()


## Caution: 1) includes deforestation and conversion of other wooded land and grassland

# [Official NFMS dataset] INPE.DETER data from 2nd August 2016 up to the 04th of April 2025
# Subsetting criteria: forest degradation classes ['CICATRIZ_DE_QUEIMADA', 'CS_DESORDENADO', 'DEGRADACAO'] and view_date until 2020-12-31
# 'CS_GEOMETRICO' excluded to align with FREL


def nbr_deter_amazon_before_2020_prep():
    deteramz = ee.FeatureCollection("projects/ee-whisp/assets/NBR/deter_amz_16apr2025")
    degradation_classes = ["CICATRIZ_DE_QUEIMADA", "CS_DESORDENADO", "DEGRADACAO"]

    # Add a formatted date field based on VIEW_DATE
    def add_formatted_date(feature):
        return feature.set("formatted_date", ee.Date(feature.get("VIEW_DATE")))

    deteramz = deteramz.map(add_formatted_date)

    deter_deg = deteramz.filter(
        ee.Filter.inList("CLASSNAME", degradation_classes)
    ).filter(ee.Filter.lt("formatted_date", ee.Date("2020-12-31")))

    deter_deg_binary = ee.Image().paint(deter_deg, 1)
    return deter_deg_binary.rename(
        "nBR_DETER_forestdegradation_Amazon_before_2020"
    ).selfMask()


################ ### NBR Disturbances after 2020:########################################
# [Official NFMS dataset] INPE PRODES data up to 2023
# Subsetting criteria: DN = [21, 22, 23, 61, 62, 63];

# the resulting dataset shows deforestation and conversion of OWL and OL up to 2020 (mostly August 2020), including residues (omission errors corrections)


def nbr_prodes_after_2020_prep():
    prodes = ee.Image("projects/ee-whisp/assets/NBR/prodes_brasil_2023")
    prodes_after_20_dn = [21, 22, 23, 61, 62, 63]
    prodes_after_20_mask = prodes.remap(
        prodes_after_20_dn, [1] * len(prodes_after_20_dn)
    )  # .eq(1)
    prodes_after_20 = prodes_after_20_mask.selfMask()
    return prodes_after_20.rename(
        "nBR_PRODES_deforestation_Brazil_after_2020"
    ).selfMask()


# %%
# [Official NFMS dataset] INPE.DETER data from 2nd August 2016 up to the 04th of April 2025
# Subsetting criteria: forest degradation classes ['CICATRIZ_DE_QUEIMADA', 'CS_DESORDENADO', 'DEGRADACAO'] and view_date from 2021-01-01 onward
# 'CS_GEOMETRICO' excluded to align with FREL
def nbr_deter_amazon_after_2020_prep():
    deteramz = ee.FeatureCollection("projects/ee-whisp/assets/NBR/deter_amz_16apr2025")
    degradation_classes = ["CICATRIZ_DE_QUEIMADA", "CS_DESORDENADO", "DEGRADACAO"]

    # Add a formatted date field based on VIEW_DATE
    def add_formatted_date(feature):
        return feature.set("formatted_date", ee.Date(feature.get("VIEW_DATE")))

    deteramz = deteramz.map(add_formatted_date)

    deter_deg = deteramz.filter(
        ee.Filter.inList("CLASSNAME", degradation_classes)
    ).filter(ee.Filter.gt("formatted_date", ee.Date("2021-01-01")))

    deter_deg_binary = ee.Image().paint(deter_deg, 1)
    return deter_deg_binary.rename(
        "nBR_DETER_forestdegradation_Amazon_after_2020"
    ).selfMask()


# ########################## NBR commodities - permanent/perennial crops in 2020:###############################
# [Official NFMS dataset] INPE/EMBRAPA TerraClass land use/cover in the Amazon biome, 2020
# OR [Official NFMS dataset] INPE/EMBRAPA TerraClass land use/cover in the Cerrado biome, 2020
# Subsetting criteria: perennial (DN=12) and semi-perennial (DN=13) crops
# the resulting dataset shows perennial and semi-perennial crops in 2020
def nbr_terraclass_amz_cer20_pc_prep():
    tcamz20 = ee.Image("projects/ee-whisp/assets/NBR/terraclass_amz_2020")
    tcamz20_pc = tcamz20.eq(12).Or(tcamz20.eq(13))
    tccer20 = ee.Image("projects/ee-whisp/assets/NBR/terraclass_cer_2020")
    tccer20_pc = tccer20.eq(12).Or(tccer20.eq(13))
    tc_pc = ee.ImageCollection([tcamz20_pc, tccer20_pc]).mosaic()
    return tc_pc.rename("nBR_INPE_TCamz_cer_perennial_2020").selfMask()


# [non-official dataset by MapBiomas multisector initiative]
# land use/cover from 1985 up to 2023, collection 9
# Subsetting criteria: 'classification_2020' = coffee (DN=46) <================== COFFEE
# the resulting dataset shows coffee area in 2020
def nbr_mapbiomasc9_cof_prep():
    mapbiomasc9_20 = ee.Image(
        "projects/mapbiomas-public/assets/brazil/lulc/collection9/mapbiomas_collection90_integration_v1"
    ).select("classification_2020")
    mapbiomasc9_20_coffee = mapbiomasc9_20.eq(46)
    return mapbiomasc9_20_coffee.rename("nBR_MapBiomas_col9_coffee_2020").selfMask()


# [non-official dataset by MapBiomas multisector initiative]
# land use/cover from 1985 up to 2023, collection 9
# Subsetting criteria: 'classification_2020' = palm oil (DN=35) <================= PALM OIL
# the resulting dataset shows palm oil area in 2020
def nbr_mapbiomasc9_po_prep():
    mapbiomasc9_20 = ee.Image(
        "projects/mapbiomas-public/assets/brazil/lulc/collection9/mapbiomas_collection90_integration_v1"
    ).select("classification_2020")
    mapbiomasc9_20_palm = mapbiomasc9_20.eq(35)
    return mapbiomasc9_20_palm.rename("nBR_MapBiomas_col9_palmoil_2020").selfMask()


# [non-official dataset by MapBiomas multisector initiative]
# land use/cover from 1985 up to 2023, collection 9
# Subsetting criteria: 'classification_2020' = other perennial crops (DN=48)
# the resulting dataset shows citrus and perennial crops other than coffee and palm oil in 2020
def nbr_mapbiomasc9_pc_prep():
    mapbiomasc9_20 = ee.Image(
        "projects/mapbiomas-public/assets/brazil/lulc/collection9/mapbiomas_collection90_integration_v1"
    ).select("classification_2020")
    mapbiomasc9_20_pc = mapbiomasc9_20.eq(35).Or(mapbiomasc9_20.eq(46))
    return mapbiomasc9_20_pc.rename("nBR_MapBiomas_col9_pc_2020").selfMask()


# ######################## NBR commodities - annual crops in 2020:##############################

# %%
# [Official NFMS dataset] INPE/EMBRAPA TerraClass land use/cover in the Amazon biome, 2020
# [Official NFMS dataset] INPE/EMBRAPA TerraClass land use/cover in the Cerrado biome, 2020
# Subsetting criteria: annual/temporary 1 cycle (DN=14) or more than 1 cycle (DN=15)
# the resulting dataset shows temporary crop in 2020
def nbr_terraclass_amz_cer20_ac_prep():
    tcamz20 = ee.Image("projects/ee-whisp/assets/NBR/terraclass_amz_2020")
    tcamz20_ac = tcamz20.eq(14).Or(tcamz20.eq(15))
    tccer20 = ee.Image("projects/ee-whisp/assets/NBR/terraclass_cer_2020")
    tccer20_ac = tccer20.eq(14).Or(tccer20.eq(15))
    tc_ac = ee.ImageCollection([tcamz20_ac, tccer20_ac]).mosaic()
    return tc_ac.rename("nBR_INPE_TCamz_cer_annual_2020").selfMask()


# [non-official dataset by MapBiomas multisector initiative]
# land use/cover from 1985 up to 2023, collection 9
# Subsetting criteria: 'classification_2020' = soybean (DN=39) <================== SOY
# the resulting dataset shows soybean plantation area in 2020
def nbr_mapbiomasc9_soy_prep():
    mapbiomasc9_20 = ee.Image(
        "projects/mapbiomas-public/assets/brazil/lulc/collection9/mapbiomas_collection90_integration_v1"
    ).select("classification_2020")
    mapbiomasc9_20_soy = mapbiomasc9_20.eq(39)
    return mapbiomasc9_20_soy.rename("nBR_MapBiomas_col9_soy_2020").selfMask()


# [non-official dataset by MapBiomas multisector initiative]
# land use/cover from 1985 up to 2023, collection 9
# Subsetting criteria: 'classification_2020' = other temporary crops (DN=41)
# Subsetting criteria: 'classification_2020' = sugar cane (DN=20)
# Subsetting criteria: 'classification_2020' = rice (DN=40)
# Subsetting criteria: 'classification_2020' = cotton (beta version, DN=62)
# the resulting dataset shows temporary crop area other than soy, includes sugar cane, rice, and cotton
def nbr_mapbiomasc9_ac_prep():
    mapbiomasc9_20 = ee.Image(
        "projects/mapbiomas-public/assets/brazil/lulc/collection9/mapbiomas_collection90_integration_v1"
    ).select("classification_2020")
    mapbiomasc9_20_ac = (
        mapbiomasc9_20.eq(41)
        .Or(mapbiomasc9_20.eq(20))
        .Or(mapbiomasc9_20.eq(40))
        .Or(mapbiomasc9_20.eq(62))
    )
    return mapbiomasc9_20_ac.rename("nBR_MapBiomas_col9_annual_crops_2020").selfMask()


# ################################### NBR commodities - pasture/livestock in 2020:##############################

# %%
# [Official NFMS dataset] INPE/EMBRAPA TerraClass land use/cover in the Amazon biome, 2020
# Subsetting criteria: BUSH/SHRUB PASTURE (DN=10) or HERBACEOUS PASTURE (DN=11)

# the resulting dataset shows 2020 pasture area in the Amazon
def nbr_terraclass_amz20_pasture_prep():
    tcamz20 = ee.Image("projects/ee-whisp/assets/NBR/terraclass_amz_2020")
    tcamz20_pasture = tcamz20.eq(10).Or(tcamz20.eq(11))
    return tcamz20_pasture.rename("nBR_INPE_TCamz_pasture_2020").selfMask()


# %%
# [Official NFMS dataset] INPE/EMBRAPA TerraClass land use/cover in the Cerrado biome, 2020
# Subsetting criteria: PASTURE (DN=11)
# the resulting dataset shows 2020 pasture area in the Cerrado


def nbr_terraclass_cer20_ac_prep():
    tccer20 = ee.Image("projects/ee-whisp/assets/NBR/terraclass_cer_2020")
    tccer20_pasture = tccer20.eq(11)
    return tccer20_pasture.rename("nBR_INPE_TCcer_pasture_2020").selfMask()


# %%
# [non-official dataset by MapBiomas multisector initiative]
# land use/cover from 1985 up to 2023, collection 9
# Subsetting criteria: 'classification_2020' = pasture (DN=15)
# the resulting dataset shows pasture area in 2020 in Brazil
def nbr_mapbiomasc9_pasture_prep():
    mapbiomasc9_20 = ee.Image(
        "projects/mapbiomas-public/assets/brazil/lulc/collection9/mapbiomas_collection90_integration_v1"
    ).select("classification_2020")
    mapbiomasc9_20_pasture = mapbiomasc9_20.eq(15)
    return mapbiomasc9_20_pasture.rename("nBR_MapBiomas_col9_pasture_2020").selfMask()


###################################################################
# nCO - Colombia


def nco_ideam_forest_2020_prep():
    ideam_forest_raw = ee.Image("projects/ee-whisp/assets/nCO/ideam_2020_geo")
    ideam_forest = ideam_forest_raw.eq(1)  # get forest class
    return ideam_forest.rename("nCO_ideam_forest_2020").selfMask()


def nco_ideam_eufo_commission_2020_prep():
    ideam_agroforest_raw = ee.Image("projects/ee-whisp/assets/nCO/ideam_2020_geo_EUFO")
    ideam_agroforest = ideam_agroforest_raw.eq(4)  # get forest class
    return ideam_agroforest.rename("nCO_ideam_eufo_commission_2020").selfMask()


# Cocoa_bnetd
def nci_ocs2020_prep():
    return (
        ee.Image("BNETD/land_cover/v1/2020")
        .select("classification")
        .eq(9)
        .rename("nCI_Cocoa_bnetd")
    ).selfMask()  # cocoa from national land cover map for Côte d'Ivoire


# nCM - Cameroon
# data from Aurelie Shapiro (FAO) working directly with country experts - info on methods and accuracy assessment to follow


def ncm_treecover_2020_prep():
    return (
        ee.Image("projects/ee-cocoacmr/assets/land_cover/CMR_TNTMMU_2020")
        .select("FNF_2020")
        .eq(1)
        .rename("nCM_Treecover_2020")
        .selfMask()
    )


# ============================================================================
# CONTEXT BANDS (Administrative boundaries and water mask)
# ============================================================================


def g_gaul_admin_code():
    """
    GAUL 2024 Level 1 administrative boundary codes (500m resolution).
    Used for spatial context and administrative aggregation.

    Returns
    -------
    ee.Image
        Image with admin codes renamed to 'admin_code' (as int32)
    """
    admin_image = ee.Image(
        "projects/ee-andyarnellgee/assets/gaul_2024_level_1_code_500m"
    )
    # Cast to int32 to ensure integer GAUL codes, then rename
    return admin_image.rename("admin_code")


def g_water_mask_prep():
    """
    Water mask from JRC/USGS combined dataset.
    Used to identify water bodies for downstream filtering and context.

    Multiplied by pixel area to get water area in hectares.

    Returns
    -------
    ee.Image
        Binary water mask image renamed to In_waterbody (will be multiplied by pixel area)
    """
    from openforis_whisp.parameters.config_runtime import water_flag

    water_mask_image = ee.Image("projects/ee-andyarnellgee/assets/water_mask_jrc_usgs")
    return water_mask_image.selfMask().rename(water_flag)


###Combining datasets


def combine_datasets(
    national_codes=None,
    validate_bands=False,
    include_context_bands=True,
    auto_recovery=False,
):
    """
    Combines datasets into a single multiband image, with fallback if assets are missing.

    Parameters
    ----------
    national_codes : list, optional
        List of ISO2 country codes to include national datasets
    validate_bands : bool, optional
        If True, validates band names with a slow .getInfo() call (default: False)
        Only enable for debugging. Normal operation relies on exception handling.
    include_context_bands : bool, optional
        If True (default), includes context bands (admin_code, water_flag) in the output.
        Set to False when using stats.py implementations that compile datasets differently.
    auto_recovery : bool, optional
        If True (default), automatically enables validate_bands when an error is detected
        during initial assembly. This allows graceful recovery from missing/broken datasets.

    Returns
    -------
    ee.Image
        Combined multiband image with all datasets (and optionally context bands)
    """
    # Step 1: Combine all main dataset images
    all_images = [ee.Image(1).rename(geometry_area_column)]
    for func in list_functions(national_codes=national_codes):
        try:
            all_images.append(func())
        except ee.EEException as e:
            print(f"Error loading image: {e}")

    img_combined = ee.Image.cat(all_images)

    # Step 2: Determine if validation needed
    should_validate = validate_bands
    if auto_recovery and not validate_bands:
        try:
            # Fast error detection: batch check main + context bands in one call
            bands_to_check = [img_combined.bandNames().get(0)]
            if include_context_bands:
                admin_image = g_gaul_admin_code()
                water_mask = g_water_mask_prep()
                bands_to_check.extend(
                    [admin_image.bandNames().get(0), water_mask.bandNames().get(0)]
                )
            ee.List(bands_to_check).getInfo()  # trigger error if any band is invalid
        except ee.EEException as e:
            print(f"Error detected, enabling recovery mode: {str(e)[:80]}...")
            should_validate = True

    # Step 3: Validate and recover if needed
    if should_validate:
        try:
            img_combined.bandNames().getInfo()  # check all bands
        except ee.EEException as e:
            print("Using valid datasets filter due to error in validation")
            valid_imgs = keep_valid_images(
                [func() for func in list_functions(national_codes=national_codes)]
            )
            all_images_retry = [ee.Image(1).rename(geometry_area_column)]
            all_images_retry.extend(valid_imgs)
            img_combined = ee.Image.cat(all_images_retry)

    # Step 4: Multiply main datasets by pixel area
    img_combined = img_combined.multiply(ee.Image.pixelArea())

    # Step 5: Add context bands (admin_code only - water mask is now in prep functions)
    if include_context_bands:
        for band_func, band_name in [
            (g_gaul_admin_code, "admin_code"),
            (g_water_mask_prep, "In_waterbody"),
        ]:
            try:
                band_img = band_func()
                if should_validate:
                    band_img.bandNames().getInfo()
                img_combined = img_combined.addBands(band_img)
            except ee.EEException as e:
                print(f"Warning: Could not add {band_name} band: {e}")

    print("Whisp multiband image compiled")
    return img_combined


######helper functions to check images
# list all functions ending with "_prep" (in the current script)
# def list_functions():
#     # Use the module's globals to get all defined functions
#     current_module = inspect.getmodule(inspect.currentframe())
#     functions = [
#         func
#         for name, func in inspect.getmembers(current_module, inspect.isfunction)
#         if name.endswith("_prep")
#     ]
#     return functions


def list_functions(national_codes=None):
    """
    Returns a list of functions that end with "_prep" and either:
    - Start with "g_" (global/regional products, excluding context bands)
    - Start with any provided national code prefix (nXX_)

    Context band functions (g_gaul_admin_code, g_water_mask_prep) are handled
    separately and excluded from this list to avoid duplication.

    Args:
        national_codes: List of ISO2 country codes (without the 'n' prefix)
    """
    # Use the module's globals to get all defined functions
    current_module = inspect.getmodule(inspect.currentframe())

    # If national_codes is None, default to an empty list
    if national_codes is None:
        national_codes = []

    # Context band functions that are handled separately
    context_functions = {"g_gaul_admin_code", "g_water_mask_prep"}

    # Create prefixes list with proper formatting ('n' + code + '_')
    allowed_prefixes = ["g_"] + [f"n{code.lower()}_" for code in national_codes]

    # Filter functions in a single pass, excluding context band functions
    functions = [
        func
        for name, func in inspect.getmembers(current_module, inspect.isfunction)
        if name.endswith("_prep")
        and any(name.startswith(prefix) for prefix in allowed_prefixes)
        and name not in context_functions
    ]

    return functions


# # IN PROGRESS - expected behaviour
# def filter_by_prefix_list(input_list=None,prefix_list=None):

#     if input_list is None:
#         print ("No function in list")
#     if prefix_list is None:
#         print ("No prefixes listed by which to filter")
#     if input_list is not None:
#         for prefix in prefix_list:
#             if element.startsWith(prefix):
#                 list.


def keep_valid_images(images):
    """Keeps only valid images."""
    valid_images = []
    for img in images:
        try:
            img.getInfo()  # This will raise an exception if the image is invalid
            valid_images.append(img)
        except ee.EEException as e:
            # logger.error(f"Invalid image: {e}")
            print(f"Invalid image: {e}")
    return valid_images


# function to check if an image is valid
def ee_image_checker(image):
    """
    Tests if the input is a valid ee.Image.

    Args:
        image: An ee.Image object.

    Returns:
        bool: True if the input is a valid ee.Image, False otherwise.
    """
    try:
        if ee.Algorithms.ObjectType(image).getInfo() == "Image":
            # Trigger some action on the image to ensure it's a valid image
            image.getInfo()  # This will raise an exception if the image is invalid
            return True
    except ee.EEException as e:
        print(f"Image validation failed with EEException: {e}")
    except Exception as e:
        print(f"Image validation failed with exception: {e}")
    return False


# preparation steps for multiband image with area per pixel values
# function for notebook environment
# user provides custom_images dict and custom_bands_info dict
def combine_custom_bands(custom_images, custom_bands_info):
    """
    Combine custom Earth Engine images into a single multiband image with area conversion.

    Returns
    -------
    ee.Image
        Combined bands converted to area values
    """
    # ... existing validation code ...

    # Step 3: Rename and combine images
    band_names = list(custom_bands_info.keys())

    # Start with first image
    custom_ee_image = custom_images[band_names[0]].rename(band_names[0])

    # Add remaining images if any
    for name in band_names[1:]:
        next_image = custom_images[name].rename(name)
        custom_ee_image = custom_ee_image.addBands(next_image)

    # Convert to area values
    custom_ee_image = custom_ee_image.multiply(ee.Image.pixelArea())

    return custom_ee_image  # Only return the image


# %%



"""
Advanced statistics processing for WHISP - concurrent and sequential endpoints.

This module provides optimized functions for processing GeoJSON FeatureCollections
with Whisp datasets using concurrent batching (for high-volume processing)
and standard sequential processing.

NOTE: This module is a transition state. The plan is to eventually merge these
functions into stats.py and replace the standard functions there as the primary
implementation, deprecating the legacy versions.

Key features:
  - whisp_stats_geojson_to_df_concurrent
  - whisp_stats_geojson_to_df_sequential (standard endpoint, sequential)
  - Proper logging at different levels (WARNING, INFO, DEBUG)
  - Progress tracking without external dependencies
  - Client-side and server-side metadata extraction options
  - Endpoint validation and warnings
"""

import ee
import pandas as pd
import geopandas as gpd
import logging
import sys
import threading
import time
import warnings
import json
import io
import os
import subprocess
from contextlib import redirect_stdout, contextmanager
from pathlib import Path
from typing import Optional, List, Dict, Any, Tuple, Union
from importlib.metadata import version as get_version
from concurrent.futures import ThreadPoolExecutor, as_completed
import tempfile

# Configure the "whisp" logger with auto-flush handler for Colab visibility
_whisp_logger = logging.getLogger("whisp")
if not _whisp_logger.handlers:
    _handler = logging.StreamHandler(sys.stdout)
    _handler.setLevel(logging.DEBUG)
    _handler.setFormatter(logging.Formatter("%(levelname)s: %(message)s"))
    # Override emit to force flush after each message for Colab
    _original_emit = _handler.emit

    def _emit_with_flush(record):
        _original_emit(record)
        sys.stdout.flush()

    _handler.emit = _emit_with_flush
    _whisp_logger.addHandler(_handler)
    _whisp_logger.setLevel(logging.INFO)
    _whisp_logger.propagate = False  # Don't propagate to root to avoid duplicates

# ============================================================================
# STDOUT/STDERR SUPPRESSION CONTEXT MANAGER (for C-level output)
# ============================================================================


@contextmanager
def suppress_c_level_output():
    """Suppress C-level stdout/stderr writes from libraries like Fiona."""
    if sys.platform == "win32":
        # Windows doesn't support dup2() reliably for STDOUT/STDERR
        # Fall back to Python-level suppression
        with redirect_stdout(io.StringIO()):
            yield
    else:
        # Unix-like systems: use file descriptor redirection
        saved_stdout = os.dup(1)
        saved_stderr = os.dup(2)
        try:
            devnull = os.open(os.devnull, os.O_WRONLY)
            os.dup2(devnull, 1)
            os.dup2(devnull, 2)
            yield
        finally:
            os.dup2(saved_stdout, 1)
            os.dup2(saved_stderr, 2)
            os.close(devnull)
            os.close(saved_stdout)
            os.close(saved_stderr)


# Suppress verbose warnings globally for this module
# Note: FutureWarnings are kept (they signal important API changes)
warnings.filterwarnings("ignore", category=UserWarning, message=".*geographic CRS.*")
warnings.simplefilter("ignore", UserWarning)
warnings.filterwarnings("ignore", category=DeprecationWarning)

# Suppress verbose logging from GeoPandas/Fiona/pyogrio
logging.getLogger("fiona").setLevel(logging.WARNING)
logging.getLogger("fiona.ogrext").setLevel(logging.WARNING)
logging.getLogger("pyogrio").setLevel(logging.WARNING)
logging.getLogger("pyogrio._io").setLevel(logging.WARNING)

from openforis_whisp.parameters.config_runtime import (
    plot_id_column,
    external_id_column,
    geometry_type_column,
    geometry_area_column,
    centroid_x_coord_column,
    centroid_y_coord_column,
    iso3_country_column,
    iso2_country_column,
    admin_1_column,
    water_flag,
    geometry_area_column_formatting,
    stats_area_columns_formatting,
    stats_percent_columns_formatting,
)
from openforis_whisp.data_conversion import (
    convert_geojson_to_ee,
    convert_ee_to_df,
    convert_ee_to_geojson,
)
from openforis_whisp.datasets import combine_datasets
from openforis_whisp.reformat import validate_dataframe_using_lookups_flexible
from openforis_whisp.stats import (
    reformat_geometry_type,
    set_point_geometry_area_to_zero,
)


# ============================================================================
# LOGGING & PROGRESS UTILITIES
# ============================================================================


def _suppress_verbose_output(max_concurrent: int = None):
    """
    Suppress verbose warnings and logging from dependencies.

    Dynamically adjusts urllib3 logger level based on max_concurrent to prevent
    "Connection pool is full" warnings during high-concurrency scenarios.

    Parameters
    ----------
    max_concurrent : int, optional
        Maximum concurrent workers. Adjusts urllib3 logging level:
        - max_concurrent <= 20: WARNING (pool rarely full)
        - max_concurrent 21-35: CRITICAL (suppress pool warnings)
        - max_concurrent >= 36: CRITICAL (maximum suppression)
    """
    warnings.filterwarnings("ignore", category=UserWarning)
    warnings.filterwarnings("ignore", category=DeprecationWarning)

    # Suppress urllib3 connection pool warnings via filters
    warnings.filterwarnings("ignore", message=".*Connection pool is full.*")
    warnings.filterwarnings("ignore", message=".*discarding connection.*")

    # Set logger levels to WARNING to suppress INFO messages
    for mod_name in [
        "openforis_whisp.reformat",
        "openforis_whisp.data_conversion",
        "geopandas",
        "fiona",
        "pyogrio._io",
        "urllib3",
    ]:
        logging.getLogger(mod_name).setLevel(logging.WARNING)

    # ALL urllib3 loggers: use CRITICAL to suppress ALL connection pool warnings
    # (these appear at WARNING level during high concurrency)
    urllib3_loggers = [
        "urllib3.connectionpool",
        "urllib3.poolmanager",
        "urllib3",
        "requests.packages.urllib3.connectionpool",
        "requests.packages.urllib3.poolmanager",
        "requests.packages.urllib3",
    ]

    for logger_name in urllib3_loggers:
        logging.getLogger(logger_name).setLevel(logging.CRITICAL)

    # Suppress warning logs specifically from reformat module during validation
    reformat_logger = logging.getLogger("openforis_whisp.reformat")
    reformat_logger.setLevel(logging.ERROR)


def _load_and_prepare_geojson(
    filepath: str, external_id_column: Optional[str] = None
) -> gpd.GeoDataFrame:
    """Load GeoJSON file and prepare for processing.

    Suppresses logging output and optionally renames external_id column.

    Parameters
    ----------
    filepath : str
        Path to GeoJSON file
    external_id_column : str, optional
        If provided, rename this column to 'external_id' immediately after loading

    Returns
    -------
    gpd.GeoDataFrame
        Loaded GeoDataFrame with external_id renamed if specified
    """
    fiona_logger = logging.getLogger("fiona")
    pyogrio_logger = logging.getLogger("pyogrio._io")
    old_fiona_level = fiona_logger.level
    old_pyogrio_level = pyogrio_logger.level
    fiona_logger.setLevel(logging.CRITICAL)
    pyogrio_logger.setLevel(logging.CRITICAL)

    try:
        with redirect_stdout(io.StringIO()):
            gdf = gpd.read_file(filepath)

        # Rename external_id column early and convert to string
        if external_id_column and external_id_column in gdf.columns:
            if external_id_column != "external_id":
                gdf = gdf.rename(
                    columns={external_id_column: "external_id"}
                )  # hard coding here to avoid confusion later
            # Convert to string to ensure consistent type throughout pipeline
            gdf["external_id"] = gdf["external_id"].astype(str)

        return gdf
    finally:
        fiona_logger.setLevel(old_fiona_level)
        pyogrio_logger.setLevel(old_pyogrio_level)


def _extract_decimal_places(format_string: str) -> int:
    """
    Extract decimal places from a format string like '%.3f'.

    Parameters
    ----------
    format_string : str
        Format string (e.g., '%.3f' → 3)

    Returns
    -------
    int
        Number of decimal places
    """
    import re

    match = re.search(r"\.(\d+)f", format_string)
    if match:
        return int(match.group(1))
    return 2  # Default to 2 decimal places


def _normalize_keep_external_columns(
    keep_external_columns: Union[bool, List[str]],
    all_columns: List[str],
    plot_id_column: str = "plotId",
) -> List[str]:
    """
    Normalize keep_external_columns parameter to a list of column names.

    Converts flexible user input (bool or list) to a concrete list of columns to keep.

    Parameters
    ----------
    keep_external_columns : bool or List[str]
        - False: keep nothing (return empty list)
        - True: keep all columns except geometry and plot_id
        - List[str]: keep specific columns (return as-is)
    all_columns : List[str]
        All available columns to choose from
    plot_id_column : str
        Name of plot ID column to exclude

    Returns
    -------
    List[str]
        Columns to keep from external (GeoJSON) data

    Examples
    --------
    >>> cols = _normalize_keep_external_columns(False, ["id", "Country", "geom"], "id")
    >>> cols
    []

    >>> cols = _normalize_keep_external_columns(True, ["id", "Country", "geom"], "id")
    >>> cols
    ['Country']

    >>> cols = _normalize_keep_external_columns(["Country"], ["id", "Country", "geom"], "id")
    >>> cols
    ['Country']
    """
    if keep_external_columns is True:
        # Keep all columns except geometry and plot_id
        return [c for c in all_columns if c not in [plot_id_column, "geometry"]]
    elif keep_external_columns is False:
        # Keep nothing
        return []
    else:
        # Use provided list (handle None case)
        return keep_external_columns or []


def _add_admin_context(
    df: pd.DataFrame, admin_code_col: str = "admin_code_median", debug: bool = False
) -> pd.DataFrame:
    """
    Join admin codes to get Country, ProducerCountry, and Admin_Level_1 information.

    Uses GAUL 2024 Level 1 administrative lookup to map admin codes to country and
    administrative region names.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame with admin_code_median column from reduceRegions
    admin_code_col : str
        Name of the admin code column (default: "admin_code_median")
    debug : bool
        If True, print detailed debugging information (default: False)

    Returns
    -------
    pd.DataFrame
        DataFrame with added Country, ProducerCountry, Admin_Level_1 columns
    """
    logger = logging.getLogger("whisp")

    # Return early if admin code column doesn't exist
    if admin_code_col not in df.columns:
        logger.debug(f"Admin code column '{admin_code_col}' not found in dataframe")
        if debug:
            print(f"DEBUG: Admin code column '{admin_code_col}' not found")
            print(f"DEBUG: Available columns: {df.columns.tolist()}")
        return df

    try:
        from openforis_whisp.parameters.lookup_gaul1_admin import lookup_dict

        if debug:
            print(f"DEBUG: Found admin_code_col '{admin_code_col}'")
            print(f"DEBUG: Sample values: {df[admin_code_col].head()}")
            print(f"DEBUG: Value types: {df[admin_code_col].dtype}")
            print(f"DEBUG: Null count: {df[admin_code_col].isna().sum()}")

        # Create lookup dataframe
        lookup_data = []
        for gaul_code, info in lookup_dict.items():
            lookup_data.append(
                {
                    "gaul1_code": gaul_code,
                    "gaul1_name": info.get("gaul1_name"),
                    "iso3_code": info.get("iso3_code"),
                    "iso2_code": info.get("iso2_code"),
                }
            )

        lookup_df = pd.DataFrame(lookup_data)

        if debug:
            print(f"DEBUG: Lookup dictionary has {len(lookup_df)} entries")
            print(f"DEBUG: Sample lookup codes: {lookup_df['gaul1_code'].head()}")

        # Prepare data for join
        df = df.copy()
        df["admin_code_for_join"] = df[admin_code_col].fillna(-9999).astype("int32")
        lookup_df["gaul1_code"] = lookup_df["gaul1_code"].astype("int32")

        if debug:
            print(
                f"DEBUG: Codes to join (first 10): {df['admin_code_for_join'].unique()[:10]}"
            )

        # Perform join
        df_joined = df.merge(
            lookup_df, left_on="admin_code_for_join", right_on="gaul1_code", how="left"
        )

        if debug:
            matched = df_joined["iso3_code"].notna().sum()
            print(f"DEBUG: Merge result - {matched}/{len(df_joined)} rows matched")
            print(f"DEBUG: Sample matched rows:")
            print(
                df_joined[
                    ["admin_code_for_join", "iso3_code", "iso2_code", "gaul1_name"]
                ].head()
            )

        # Rename columns to match output schema
        df_joined = df_joined.rename(
            columns={
                "iso3_code": iso3_country_column,  # 'Country'
                "iso2_code": iso2_country_column,  # 'ProducerCountry'
                "gaul1_name": admin_1_column,  # 'Admin_Level_1'
            }
        )

        # Drop temporary columns
        df_joined = df_joined.drop(
            columns=["admin_code_for_join", "gaul1_code"], errors="ignore"
        )

        logger.debug(
            f"Admin context added: {iso3_country_column}, {iso2_country_column}, {admin_1_column}"
        )
        return df_joined

    except ImportError:
        logger.warning(
            "Could not import GAUL lookup dictionary - admin context not added"
        )
        if debug:
            print("DEBUG: ImportError - could not load lookup dictionary")
        return df
    except Exception as e:
        logger.warning(f"Error adding admin context: {e}")
        if debug:
            print(f"DEBUG: Exception in _add_admin_context: {e}")
            import traceback

            traceback.print_exc()
        return df


def join_admin_codes(
    df: pd.DataFrame, lookup_dict: Dict, id_col: str = "admin_code_median"
) -> pd.DataFrame:
    """
    Join admin codes to DataFrame using a lookup dictionary.

    Converts the admin code column to integer and performs a left join with
    the lookup dictionary to add Country, ProducerCountry, and Admin_Level_1.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame with admin code column
    lookup_dict : dict
        Dictionary mapping GAUL codes to admin info (iso3_code, iso2_code, gaul1_name)
    id_col : str
        Name of the admin code column (default: "admin_code_median")

    Returns
    -------
    pd.DataFrame
        DataFrame with added Country, ProducerCountry, Admin_Level_1 columns
    """
    logger = logging.getLogger("whisp")

    # Return early if admin code column doesn't exist
    if id_col not in df.columns:
        logger.debug(f"Admin code column '{id_col}' not found in dataframe")
        return df

    try:
        # Create lookup dataframe
        lookup_data = []
        for gaul_code, info in lookup_dict.items():
            lookup_data.append(
                {
                    "gaul1_code": gaul_code,
                    "gaul1_name": info.get("gaul1_name"),
                    "iso3_code": info.get("iso3_code"),
                    "iso2_code": info.get("iso2_code"),
                }
            )

        lookup_df = pd.DataFrame(lookup_data)

        # Prepare data for join
        df = df.copy()
        # Round to nearest integer (handles float values from EE reducers)
        df["admin_code_for_join"] = df[id_col].fillna(-9999).astype("int32")
        lookup_df["gaul1_code"] = lookup_df["gaul1_code"].astype("int32")

        # Perform join
        df_joined = df.merge(
            lookup_df, left_on="admin_code_for_join", right_on="gaul1_code", how="left"
        )

        # Rename columns to match output schema
        df_joined = df_joined.rename(
            columns={
                "iso3_code": iso3_country_column,  # 'Country'
                "iso2_code": iso2_country_column,  # 'ProducerCountry'
                "gaul1_name": admin_1_column,  # 'Admin_Level_1'
            }
        )

        # Drop temporary columns
        df_joined = df_joined.drop(
            columns=["admin_code_for_join", "gaul1_code"], errors="ignore"
        )

        # Fill NaN values with "Unknown" and "not found" for features outside admin boundaries
        # (e.g., points in the ocean or international waters)
        df_joined[iso3_country_column] = df_joined[iso3_country_column].fillna(
            "Unknown"
        )
        df_joined[iso2_country_column] = df_joined[iso2_country_column].fillna(
            "not found"
        )
        df_joined[admin_1_column] = df_joined[admin_1_column].fillna("Unknown")

        logger.debug(
            f"Admin codes joined: {iso3_country_column}, {iso2_country_column}, {admin_1_column}"
        )
        return df_joined

    except Exception as e:
        logger.warning(f"Error joining admin codes: {e}")
        return df


def _format_time(seconds: float) -> str:
    """Format seconds as human-readable string."""
    if seconds < 60:
        return f"{seconds:.0f}s"
    elif seconds < 3600:
        mins = seconds / 60
        return f"{mins:.1f}m"
    else:
        hours = seconds / 3600
        return f"{hours:.1f}h"


def _get_progress_milestones(total_features: int) -> set:
    """
    Get progress milestones based on dataset size.

    Parameters
    ----------
    total_features : int
        Total number of features being processed

    Returns
    -------
    set
        Set of percentage milestones to show
    """
    # Set milestones based on feature count
    if total_features < 250:
        return set(range(20, 101, 20))  # Every 20%: {20, 40, 60, 80, 100}
    elif total_features < 1000:
        return set(range(10, 101, 10))  # Every 10%
    elif total_features < 10000:
        return set(range(5, 101, 5))  # Every 5%
    elif total_features < 50000:
        return set(range(2, 101, 2))  # Every 2%
    else:
        return set(range(1, 101))  # Every 1%


def _log_progress(
    completed: int,
    total: int,
    milestones: set,
    shown_milestones: set,
    start_time: float,
    logger: logging.Logger,
) -> None:
    """
    Log progress at milestone percentages.

    Parameters
    ----------
    completed : int
        Number of batches completed
    total : int
        Total number of batches
    milestones : set
        Set of percentage milestones to show
    shown_milestones : set
        Set of milestones already shown (modified in place)
    start_time : float
        Start time from time.time()
    logger : logging.Logger
        Logger for output
    """
    percent = int((completed / total) * 100)

    # Check for new milestones reached
    for milestone in sorted(milestones):
        if percent >= milestone and milestone not in shown_milestones:
            shown_milestones.add(milestone)

            # Calculate time metrics
            elapsed = time.time() - start_time
            rate = completed / elapsed if elapsed > 0 else 0
            remaining_items = total - completed

            # Calculate ETA with padding for overhead (loading, joins, etc.)
            # Don't show ETA until we have some samples (at least 5% complete)
            if rate > 0 and completed >= max(5, total * 0.05):
                eta_seconds = (remaining_items / rate) * 1.15  # Add 15% padding
            else:
                eta_seconds = 0

            # Format time strings
            eta_str = _format_time(eta_seconds) if eta_seconds > 0 else "calculating..."
            elapsed_str = _format_time(elapsed)

            # Build progress message
            msg = f"Progress: {completed:,}/{total:,} batches ({percent}%)"
            if percent < 100:
                msg += f" | Elapsed: {elapsed_str} | ETA: {eta_str}"
            else:
                msg += f" | Total time: {elapsed_str}"

            logger.info(msg)


# ============================================================================
# ENDPOINT VALIDATION
# ============================================================================


def check_ee_endpoint(endpoint_type: str = "high-volume") -> bool:
    """
    Check if Earth Engine is using the correct endpoint.

    Parameters
    ----------
    endpoint_type : str
        Expected endpoint type: "high-volume" or "standard"

    Returns
    -------
    bool
        True if using expected endpoint, False otherwise
    """
    api_url = str(ee.data._cloud_api_base_url)

    if endpoint_type == "high-volume":
        return "highvolume" in api_url.lower()
    elif endpoint_type == "standard":
        return "highvolume" not in api_url.lower()
    else:
        return False


def validate_ee_endpoint(endpoint_type: str = "high-volume", raise_error: bool = True):
    """
    Validate Earth Engine endpoint and warn/error if incorrect.

    Parameters
    ----------
    endpoint_type : str
        Expected endpoint type
    raise_error : bool
        If True, raise error if incorrect endpoint; if False, warn

    Raises
    ------
    RuntimeError
        If incorrect endpoint and raise_error=True
    """
    if not check_ee_endpoint(endpoint_type):
        if endpoint_type == "high-volume":
            msg = (
                "# Concurrent mode requires the HIGH-VOLUME endpoint. To change endpoint run:\n"
                "ee.Reset()\n"
                "ee.Initialize(project=gee_project_name, opt_url='https://earthengine-highvolume.googleapis.com')\n"
                "# where gee_project_name is your GEE project (necessary in Colab)"
            )
        else:  # standard endpoint
            msg = (
                "Sequential mode requires the STANDARD endpoint. To change endpoint run:\n"
                "ee.Reset()\n"
                "ee.Initialize(project=gee_project_name)\n"
                "# where gee_project_name is your GEE project (necessary in Colab)"
            )

        if raise_error:
            raise RuntimeError(msg)
        else:
            logging.warning(msg)


# ============================================================================
# METADATA EXTRACTION (CLIENT & SERVER SIDE)
# ============================================================================


def extract_centroid_and_geomtype_client(
    gdf: gpd.GeoDataFrame,
    x_col: str = None,
    y_col: str = None,
    type_col: str = None,
    external_id_column: str = None,
    return_attributes_only: bool = True,
) -> pd.DataFrame:
    """
    Extract centroid coordinates and geometry type using GeoPandas (client-side).

    Parameters
    ----------
    gdf : gpd.GeoDataFrame
        Input GeoDataFrame
    x_col : str, optional
        Column name for centroid x. Defaults to config value
    y_col : str, optional
        Column name for centroid y. Defaults to config value
    type_col : str, optional
        Column name for geometry type. Defaults to config value
    external_id_column: : str, optional
        Name of external ID column to preserve
    return_attributes_only : bool
        If True, return only attribute columns (no geometry)

    Returns
    -------
    pd.DataFrame or gpd.GeoDataFrame
        DataFrame/GeoDataFrame with centroid and geometry type columns
    """
    x_col = x_col or centroid_x_coord_column
    y_col = y_col or centroid_y_coord_column
    type_col = type_col or geometry_type_column

    gdf = gdf.copy()

    # Extract centroid coordinates (suppressing geographic CRS warning from Shapely)
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning)
        warnings.simplefilter("ignore", UserWarning)  # Additional suppression
        centroid_points = gdf.geometry.centroid

    gdf[x_col] = centroid_points.x.round(6)
    gdf[y_col] = centroid_points.y.round(6)
    gdf[type_col] = gdf.geometry.geom_type

    if return_attributes_only:
        # Build column list starting with merge keys
        cols = []

        # Always include __row_id__ first if present (needed for row-level merging)
        if "__row_id__" in gdf.columns:
            cols.append("__row_id__")

        # Always include plot_id_column if present (needed for merging batches)
        if plot_id_column in gdf.columns:
            cols.append(plot_id_column)

        # Include external_id if it exists (already renamed during load)
        if (
            external_id_column
            and "external_id" in gdf.columns
            and "external_id" not in cols
        ):
            cols.append("external_id")

        # Always include metadata columns (centroid, geometry type)
        cols.extend([x_col, y_col, type_col])

        # Remove any duplicates while preserving order
        cols = list(dict.fromkeys(cols))

        return gdf[cols].reset_index(drop=True)

    return gdf


def extract_centroid_and_geomtype_server(
    fc: ee.FeatureCollection,
    x_col: str = None,
    y_col: str = None,
    type_col: str = None,
    max_error: float = 1.0,
) -> ee.FeatureCollection:
    """
    Extract centroid coordinates and geometry type using Earth Engine (server-side).

    Parameters
    ----------
    fc : ee.FeatureCollection
        Input FeatureCollection
    x_col : str, optional
        Column name for centroid x
    y_col : str, optional
        Column name for centroid y
    type_col : str, optional
        Column name for geometry type
    max_error : float
        Maximum error for centroid calculation (meters)

    Returns
    -------
    ee.FeatureCollection
        FeatureCollection with centroid and geometry type properties
    """
    x_col = x_col or centroid_x_coord_column
    y_col = y_col or centroid_y_coord_column
    type_col = type_col or geometry_type_column

    def add_metadata(feature):
        centroid = feature.geometry().centroid(max_error)
        coords = centroid.coordinates()
        x = ee.Number(coords.get(0)).multiply(1e6).round().divide(1e6)
        y = ee.Number(coords.get(1)).multiply(1e6).round().divide(1e6)
        return feature.set({x_col: x, y_col: y, type_col: feature.geometry().type()})

    return fc.map(add_metadata)


# ============================================================================
# BATCH PROCESSING UTILITIES
# ============================================================================


def batch_geodataframe(
    gdf: gpd.GeoDataFrame,
    batch_size: int,
) -> List[gpd.GeoDataFrame]:
    """
    Split a GeoDataFrame into batches.

    Parameters
    ----------
    gdf : gpd.GeoDataFrame
        Input GeoDataFrame
    batch_size : int
        Size of each batch

    Returns
    -------
    List[gpd.GeoDataFrame]
        List of batch GeoDataFrames
    """
    batches = []
    for i in range(0, len(gdf), batch_size):
        batches.append(gdf.iloc[i : i + batch_size].copy())
    return batches


def convert_batch_to_ee(batch_gdf: gpd.GeoDataFrame) -> ee.FeatureCollection:
    """
    Convert a batch GeoDataFrame to EE FeatureCollection efficiently.

    OPTIMIZATION: Passes GeoDataFrame directly to convert_geojson_to_ee to preserve CRS.
    This ensures proper coordinate system handling and reprojection to WGS84 if needed.

    Preserves the __row_id__ column if present so it can be retrieved after processing.

    IMPORTANT: Drops external_id column before sending to EE to enable query caching.
    external_id is user metadata that's not needed for EE computation. Including it
    breaks EE's caching mechanism since each unique external_id creates a different query.

    Parameters
    ----------
    batch_gdf : gpd.GeoDataFrame
        Input batch (should have __row_id__ column)

    Returns
    -------
    ee.FeatureCollection
        EE FeatureCollection with __row_id__ as a feature property (no external_id)
    """
    # Drop external_id before sending to EE to enable caching
    # (external_id is preserved separately on client side for merging)
    batch_for_ee = batch_gdf.copy()
    if "external_id" in batch_for_ee.columns:
        batch_for_ee = batch_for_ee.drop(columns=["external_id"])

    # Pass GeoDataFrame directly to preserve CRS metadata
    # convert_geojson_to_ee will handle:
    # - CRS detection and conversion to WGS84 if needed
    # - Data type sanitization (datetime, object columns)
    # - Geometry validation and Z-coordinate stripping

    fc = convert_geojson_to_ee(batch_for_ee, enforce_wgs84=True, strip_z_coords=True)

    # If __row_id__ is in the original GeoDataFrame, it will be preserved
    # as a feature property in the GeoJSON and thus in the EE FeatureCollection
    return fc


def clean_geodataframe(
    gdf: gpd.GeoDataFrame,
    remove_nulls: bool = False,
    repair_geometries: bool = False,
    logger: logging.Logger = None,
) -> gpd.GeoDataFrame:
    """
    Validate and clean GeoDataFrame geometries.

    Parameters
    ----------
    gdf : gpd.GeoDataFrame
        Input GeoDataFrame
    remove_nulls : bool
        Remove null geometries. Defaults to False to preserve data integrity.
        Set to True only if you explicitly want to drop rows with null geometries.
    repair_geometries : bool
        Repair invalid geometries using Shapely's make_valid(). Defaults to False to preserve
        original geometries. Set to True only if you want to automatically repair invalid geometries.
    logger : logging.Logger, optional
        Logger for output

    Returns
    -------
    gpd.GeoDataFrame
        Cleaned GeoDataFrame
    """
    logger = logger or logging.getLogger("whisp")

    if remove_nulls:
        null_count = gdf.geometry.isna().sum()
        if null_count > 0:
            logger.warning(f"Removing {null_count} null geometries")
            gdf = gdf[~gdf.geometry.isna()].copy()

    if repair_geometries:
        valid_count = gdf.geometry.is_valid.sum()
        invalid_count = len(gdf) - valid_count
        if invalid_count > 0:
            logger.warning(f"Repairing {invalid_count} invalid geometries")
            from shapely.validation import make_valid

            gdf = gdf.copy()
            gdf["geometry"] = gdf["geometry"].apply(
                lambda g: make_valid(g) if g and not g.is_valid else g
            )

    logger.debug(f"Validation complete: {len(gdf):,} geometries ready")
    return gdf


# ============================================================================
# AUDIT TRAIL HELPER
# ============================================================================


def _add_geometry_audit_trail(
    df_validated: pd.DataFrame,
    input_geojson_filepath: str,
    gdf_original_geoms: gpd.GeoDataFrame = None,
    logger: logging.Logger = None,
) -> pd.DataFrame:
    """
    Add original input geometries as geo_original column for audit trail.

    Parameters
    ----------
    df_validated : pd.DataFrame
        Validated DataFrame to add audit trail to
    input_geojson_filepath : str
        Path to original GeoJSON file
    gdf_original_geoms : gpd.GeoDataFrame, optional
        Pre-loaded original geometries (to avoid reloading)
    logger : logging.Logger, optional
        Logger for output

    Returns
    -------
    pd.DataFrame
        DataFrame with geo_original column added
    """
    import json
    from shapely.geometry import mapping

    logger = logger or logging.getLogger("whisp")

    try:
        # Load original geometries if not provided
        if gdf_original_geoms is None:
            logger.warning("Original geometries not pre-loaded, loading now...")
            gdf_original_geoms = _load_and_prepare_geojson(input_geojson_filepath)

        # Create DataFrame with plotId and geo_original
        df_original_geom = pd.DataFrame(
            {
                "plotId": df_validated["plotId"].values[: len(gdf_original_geoms)],
                "geo_original": gdf_original_geoms["geometry"].apply(
                    lambda g: json.dumps(mapping(g)) if g is not None else None
                ),
            }
        )

        # Merge original geometries back
        df_result = df_validated.merge(df_original_geom, on="plotId", how="left")
        logger.info("Audit trail added: geo_original column")
        return df_result

    except Exception as e:
        logger.warning(f"Error adding audit trail: {e}")
        # Return original DataFrame if audit trail fails
        return df_validated


# ============================================================================
# BATCH RETRY HELPER - DEPRECATED (removed due to semaphore deadlock issues)
# ============================================================================
# Note: Retry logic via sub-batching has been removed. Instead, use fail-fast
# approach: when a batch fails, reduce batch_size parameter and retry manually.
# This avoids semaphore deadlocks and provides clearer error messages.


# ============================================================================
# EE PROCESSING WITH RETRY LOGIC
# ============================================================================


def process_ee_batch(
    fc: ee.FeatureCollection,
    whisp_image: ee.Image,
    reducer: ee.Reducer,
    batch_idx: int,
    max_retries: int = 3,
    logger: logging.Logger = None,
) -> pd.DataFrame:
    """
    Process an EE FeatureCollection with automatic retry logic.

    Parameters
    ----------
    fc : ee.FeatureCollection
        Input FeatureCollection
    whisp_image : ee.Image
        Image containing bands to reduce
    reducer : ee.Reducer
        Reducer to apply
    batch_idx : int
        Batch index (for logging)
    max_retries : int
        Maximum retry attempts
    logger : logging.Logger, optional
        Logger for output

    Returns
    -------
    pd.DataFrame
        Results as DataFrame

    Raises
    ------
    RuntimeError
        If processing fails after all retries
    """
    logger = logger or logging.getLogger("whisp")

    for attempt in range(max_retries):
        try:
            results = whisp_image.reduceRegions(
                collection=fc,
                reducer=reducer,
                scale=10,
            )
            df = convert_ee_to_df(results)

            # Ensure plot_id_column is present for merging
            # It should come from the feature properties (added before EE processing)
            if plot_id_column not in df.columns:
                logger.warning(
                    f"Batch {batch_idx + 1}: plotId column DROPPED by EE. "
                    f"Regenerating with 1-indexed range. "
                    f"Columns from EE: {list(df.columns)}"
                )
                # Use 1-indexed range to match client-side assignment
                df[plot_id_column] = [str(i) for i in range(1, len(df) + 1)]

            # Ensure plotId is string type (consistent with creation)
            if plot_id_column in df.columns:
                df[plot_id_column] = df[plot_id_column].astype(str)

            # Ensure all column names are strings (fixes pandas .str accessor issues)
            df.columns = df.columns.astype(str)

            return df

        except ee.EEException as e:
            error_msg = str(e)

            if "Quota" in error_msg or "limit" in error_msg.lower():
                if attempt < max_retries - 1:
                    wait_time = min(30, 2**attempt)
                    logger.warning(
                        f"Batch {batch_idx + 1}: Rate limited, waiting {wait_time}s..."
                    )
                    time.sleep(wait_time)
                else:
                    raise RuntimeError(f"Batch {batch_idx + 1}: Quota exhausted")

            elif "timeout" in error_msg.lower():
                if attempt < max_retries - 1:
                    wait_time = min(15, 2**attempt)
                    logger.warning(
                        f"Batch {batch_idx + 1}: Timeout, retrying in {wait_time}s..."
                    )
                    time.sleep(wait_time)
                else:
                    raise

            else:
                if attempt < max_retries - 1:
                    wait_time = min(5, 2**attempt)
                    time.sleep(wait_time)
                else:
                    raise

        except Exception as e:
            if attempt < max_retries - 1:
                time.sleep(min(5, 2**attempt))
            else:
                raise RuntimeError(f"Batch {batch_idx + 1}: {str(e)}")

    raise RuntimeError(f"Batch {batch_idx + 1}: Failed after {max_retries} attempts")


# ============================================================================
# CONCURRENT PROCESSING FUNCTIONS
# ============================================================================


def whisp_stats_geojson_to_df_concurrent(
    input_geojson_filepath: str,
    external_id_column: str = None,
    national_codes: List[str] = None,
    unit_type: str = "ha",
    whisp_image: ee.Image = None,
    custom_bands: Dict[str, Any] = None,
    batch_size: int = 10,
    max_concurrent: int = 20,
    validate_geometries: bool = True,
    max_retries: int = 3,
    add_metadata_server: bool = False,
    logger: logging.Logger = None,
    # Format parameters (auto-detect from config if not provided)
    decimal_places: int = None,
) -> pd.DataFrame:
    """
    Process GeoJSON concurrently to compute Whisp statistics with automatic formatting.

    Uses high-volume endpoint and concurrent batching. Client-side metadata
    extraction is always applied; optionally add server-side metadata too.
    Automatically formats output (converts units, removes noise columns, etc.).

    Parameters
    ----------
    input_geojson_filepath : str
        Path to input GeoJSON file
    external_id_column : str, optional
        Column name for external IDs
    national_codes : List[str], optional
        ISO2 codes for national datasets
    unit_type : str
        "ha" or "percent"
    whisp_image : ee.Image, optional
        Pre-combined image (created with combine_datasets if None)
    custom_bands : Dict[str, Any], optional
        Custom band information
    batch_size : int
        Features per batch
    max_concurrent : int
        Maximum concurrent EE calls
    validate_geometries : bool
        Validate and clean geometries
    max_retries : int
        Retry attempts per batch
    add_metadata_server : bool
        Add metadata server-side (in addition to client-side)
    logger : logging.Logger, optional
        Logger for output
    decimal_places : int, optional
        Decimal places for formatting. If None, auto-detects from config.

    Returns
    -------
    pd.DataFrame
        Formatted results DataFrame with Whisp statistics
    """
    from openforis_whisp.reformat import format_stats_dataframe

    logger = logger or logging.getLogger("whisp")

    # Suppress verbose output from dependencies (dynamically adjust based on max_concurrent)
    _suppress_verbose_output(max_concurrent=max_concurrent)

    # Auto-detect decimal places from config if not provided
    if decimal_places is None:
        decimal_places = _extract_decimal_places(stats_area_columns_formatting)
        logger.debug(f"Using decimal_places={decimal_places} from config")

    # Validate endpoint
    validate_ee_endpoint("high-volume", raise_error=True)

    # Load GeoJSON with output suppressed (external_id_column renamed to 'external_id' if provided)
    gdf = _load_and_prepare_geojson(
        input_geojson_filepath, external_id_column=external_id_column
    )
    logger.info(f"Loaded {len(gdf):,} features")

    # Validate external_id if provided (lightweight client-side check)
    # Note: external_id_column already renamed to 'external_id' during load
    if external_id_column and "external_id" not in gdf.columns:
        # Exclude geometry column from available columns list
        available_cols = [c for c in gdf.columns if c != gdf.geometry.name]
        raise ValueError(
            f"Column '{external_id_column}' not found in GeoJSON properties. "
            f"Available columns: {available_cols}"
        )

    # Check completeness of external_id (warn if nulls exist)
    if external_id_column and "external_id" in gdf.columns:
        null_count = gdf["external_id"].isna().sum()
        if null_count > 0:
            null_pct = (null_count / len(gdf)) * 100
            logger.warning(
                f"Column 'external_id' (from '{external_id_column}') has {null_count:,} null values ({null_pct:.1f}% of {len(gdf):,} features). "
                f"These features may have missing external IDs in output."
            )

    if validate_geometries:
        gdf = clean_geodataframe(
            gdf, remove_nulls=False, repair_geometries=False, logger=logger
        )

    # Add stable plotIds for merging (starting from 1, not 0)
    gdf[plot_id_column] = [str(i) for i in range(1, len(gdf) + 1)]

    # Strip unnecessary properties before sending to EE
    # Keep only: geometry, plot_id_column, and external_id
    # This prevents duplication of GeoJSON properties in EE results
    keep_cols = ["geometry", plot_id_column]
    if (
        external_id_column and "external_id" in gdf.columns
    ):  # Already renamed during load
        keep_cols.append("external_id")

    gdf_for_ee = gdf[keep_cols].copy()

    # CRITICAL: Convert external_id to string (both plotId and external_id are now strings)
    if external_id_column and "external_id" in gdf_for_ee.columns:
        gdf_for_ee["external_id"] = gdf_for_ee["external_id"].astype(str)
        logger.debug(f"Converted external_id column to string type")

    logger.debug(f"Stripped GeoJSON to essential columns: {keep_cols}")

    # Create image if not provided
    if whisp_image is None:
        logger.debug("Creating Whisp image...")
        # Suppress print statements from combine_datasets
        with redirect_stdout(io.StringIO()):
            try:
                # First try without validation
                whisp_image = combine_datasets(
                    national_codes=national_codes, validate_bands=False
                )
            except Exception as e:
                logger.warning(
                    f"First attempt failed: {str(e)[:100]}. Retrying with validate_bands=True..."
                )
                # Retry with validation to catch and fix bad bands
                whisp_image = combine_datasets(
                    national_codes=national_codes, validate_bands=True
                )

    # Create reducer
    reducer = ee.Reducer.sum().combine(ee.Reducer.median(), sharedInputs=True)

    # Batch the data
    batches = batch_geodataframe(gdf_for_ee, batch_size)
    logger.info(
        f"Processing {len(gdf_for_ee):,} features in {len(batches)} batches (concurrent mode)..."
    )

    # Setup semaphore for EE concurrency control
    ee_semaphore = threading.BoundedSemaphore(max_concurrent)

    # Progress tracking setup
    progress_lock = threading.Lock()
    completed_batches = 0
    milestones = _get_progress_milestones(len(gdf_for_ee))
    shown_milestones = set()
    start_time = time.time()

    results = []

    def process_batch(
        batch_idx: int, batch: gpd.GeoDataFrame
    ) -> Tuple[int, pd.DataFrame, pd.DataFrame]:
        """Process one batch: server EE work + client metadata."""
        with ee_semaphore:
            # Server-side: convert to EE, optionally add metadata, reduce
            fc = convert_batch_to_ee(batch)
            if add_metadata_server:
                fc = extract_centroid_and_geomtype_server(fc)
            df_server = process_ee_batch(
                fc, whisp_image, reducer, batch_idx, max_retries, logger
            )

            # Client-side: extract metadata using GeoPandas
            df_client = extract_centroid_and_geomtype_client(
                batch,
                external_id_column=external_id_column,
                return_attributes_only=True,
            )

        return batch_idx, df_server, df_client

    # Process batches with thread pool
    pool_workers = max(2 * max_concurrent, max_concurrent + 2)

    # Track if we had errors that suggest bad bands
    batch_errors = []

    # Suppress fiona logging during batch processing (threads create new loggers)
    fiona_logger = logging.getLogger("fiona")
    pyogrio_logger = logging.getLogger("pyogrio._io")
    old_fiona_level = fiona_logger.level
    old_pyogrio_level = pyogrio_logger.level
    fiona_logger.setLevel(logging.CRITICAL)
    pyogrio_logger.setLevel(logging.CRITICAL)

    try:
        # Don't suppress stdout here - we want progress messages to show in Colab
        with ThreadPoolExecutor(max_workers=pool_workers) as executor:
            futures = {
                executor.submit(process_batch, i, batch): i
                for i, batch in enumerate(batches)
            }

            # Track which batches failed for retry
            batch_map = {i: batch for i, batch in enumerate(batches)}
            batch_futures = {future: i for future, i in futures.items()}

            for future in as_completed(futures):
                batch_idx = batch_futures[future]
                try:
                    batch_idx, df_server, df_client = future.result()

                    # Merge server and client results
                    if plot_id_column not in df_server.columns:
                        logger.warning(
                            f"Batch {batch_idx + 1} (concurrent merge): plotId DROPPED by EE. "
                            f"Regenerating. Columns from EE: {list(df_server.columns)}"
                        )
                        df_server[plot_id_column] = pd.array(
                            range(1, len(df_server) + 1), dtype="Int64"
                        )
                    else:
                        df_server[plot_id_column] = df_server[plot_id_column].astype(
                            str
                        )

                    # Ensure plotId is string in client data too
                    if plot_id_column in df_client.columns:
                        df_client[plot_id_column] = df_client[plot_id_column].astype(
                            str
                        )

                    # Keep all EE statistics from server (all columns with _sum and _median suffixes)
                    # These are the actual EE processing results
                    df_server_clean = df_server.copy()

                    # Drop external_id from df_server if it exists (already in df_client)
                    if "external_id" in df_server_clean.columns:
                        df_server_clean = df_server_clean.drop(columns=["external_id"])

                    # Keep external metadata: plot_id, external_id, geometry, geometry type, and centroids from client
                    # (formatted wrapper handles keep_external_columns parameter)
                    keep_external_columns = [plot_id_column]
                    if external_id_column and "external_id" in df_client.columns:
                        keep_external_columns.append("external_id")
                    if "geometry" in df_client.columns:
                        keep_external_columns.append("geometry")
                    # Keep geometry type column (Geometry_type)
                    if geometry_type_column in df_client.columns:
                        keep_external_columns.append(geometry_type_column)
                    # Also keep centroid columns (Centroid_lon, Centroid_lat)
                    centroid_cols = [
                        c for c in df_client.columns if c.startswith("Centroid_")
                    ]
                    keep_external_columns.extend(centroid_cols)

                    df_client_clean = df_client[
                        [c for c in keep_external_columns if c in df_client.columns]
                    ]
                    # Don't drop duplicates - we need one row per feature (one per plot_id)
                    # Each plot_id should have exactly one row with its metadata

                    merged = df_server_clean.merge(
                        df_client_clean,
                        on=plot_id_column,
                        how="left",
                        suffixes=("_ee", "_client"),
                    )
                    results.append(merged)

                    # Update progress
                    with progress_lock:
                        completed_batches += 1
                        _log_progress(
                            completed_batches,
                            len(batches),
                            milestones,
                            shown_milestones,
                            start_time,
                            logger,
                        )

                except Exception as e:
                    # Batch failed - fail fast with clear guidance
                    error_msg = str(e)
                    logger.error(f"Batch {batch_idx} failed: {error_msg[:100]}")
                    logger.debug(f"Full error: {error_msg}")

                    # Get original batch for error reporting
                    original_batch = batch_map[batch_idx]

                    # Add to batch errors for final reporting
                    batch_errors.append((batch_idx, original_batch, error_msg))
    except (KeyboardInterrupt, SystemExit) as interrupt:
        logger.warning("Processing interrupted by user")
        raise interrupt
    finally:
        # Restore logger levels
        fiona_logger.setLevel(old_fiona_level)
        pyogrio_logger.setLevel(old_pyogrio_level)

    # Log completion
    total_time = time.time() - start_time
    time_str = _format_time(total_time)
    logger.info(
        f"Processing complete: {completed_batches:,}/{len(batches):,} batches in {time_str}"
    )

    # If we have batch errors after retry attempts, fail the entire process
    if batch_errors:
        total_failed_rows = sum(len(batch) for _, batch, _ in batch_errors)
        failed_batch_indices = [str(idx) for idx, _, _ in batch_errors]

        # Format detailed error information for debugging
        error_details_list = []
        for idx, batch, msg in batch_errors:
            error_details_list.append(f"  Batch {idx} ({len(batch)} features): {msg}")
        error_details = "\n".join(error_details_list)

        # Analyze error patterns for debugging hints
        error_patterns = {
            "memory": any("memory" in msg.lower() for _, _, msg in batch_errors),
            "request_size": any(
                keyword in msg.lower()
                for _, _, msg in batch_errors
                for keyword in ["too large", "10mb", "payload", "size limit"]
            ),
            "quota": any("quota" in msg.lower() for _, _, msg in batch_errors),
            "timeout": any("timeout" in msg.lower() for _, _, msg in batch_errors),
        }

        # Build helpful suggestions based on error patterns
        suggestions = []
        if error_patterns["memory"]:
            suggestions.append(
                f"  • Reduce batch_size parameter (currently: {batch_size}). Try: batch_size=5 or lower"
            )
        if error_patterns["request_size"]:
            suggestions.append(
                "  • Request payload too large: reduce batch_size or simplify input geometries"
            )
        if error_patterns["quota"]:
            suggestions.append("  • Earth Engine quota exceeded: wait and retry later")
        if error_patterns["timeout"]:
            suggestions.append(
                "  • Processing timeout: reduce batch_size or simplify input geometries"
            )

        suggestions_text = (
            "\nDebugging hints:\n" + "\n".join(suggestions) if suggestions else ""
        )

        raise RuntimeError(
            f"Failed to process {len(batch_errors)} batch(es):\n"
            f"\n{error_details}\n"
            f"\nTotal rows affected: {total_failed_rows}\n"
            f"{suggestions_text}\n"
            f"Please reduce batch_size and try again."
        )

    # Check if we should retry with validation due to band errors (legacy band error handling)
    if not results:
        # All batches failed - likely a bad band issue
        is_band_error = any(
            keyword in str(batch_errors)
            for keyword in ["Image.load", "asset", "not found", "does not exist"]
        )

        if is_band_error:
            logger.warning(
                "Detected potential bad band error. Retrying with validate_bands=True..."
            )
            try:
                with redirect_stdout(io.StringIO()):
                    whisp_image = combine_datasets(
                        national_codes=national_codes, validate_bands=True
                    )
                logger.info(
                    "Image recreated with validation. Retrying batch processing..."
                )

                # Retry batch processing with validated image
                results = []
                retry_completed = 0
                retry_shown = set()
                retry_start = time.time()

                # Suppress fiona logging during batch processing (threads create new loggers)
                fiona_logger = logging.getLogger("fiona")
                pyogrio_logger = logging.getLogger("pyogrio._io")
                old_fiona_level = fiona_logger.level
                old_pyogrio_level = pyogrio_logger.level
                fiona_logger.setLevel(logging.CRITICAL)
                pyogrio_logger.setLevel(logging.CRITICAL)

                try:
                    with ThreadPoolExecutor(max_workers=pool_workers) as executor:
                        futures = {
                            executor.submit(process_batch, i, batch): i
                            for i, batch in enumerate(batches)
                        }

                        for future in as_completed(futures):
                            try:
                                batch_idx, df_server, df_client = future.result()
                                if plot_id_column not in df_server.columns:
                                    # Use 1-indexed range to match client-side assignment
                                    df_server[plot_id_column] = range(
                                        1, len(df_server) + 1
                                    )
                                merged = df_server.merge(
                                    df_client,
                                    on=plot_id_column,
                                    how="left",
                                    suffixes=("", "_client"),
                                )
                                results.append(merged)

                                # Update retry progress
                                with progress_lock:
                                    retry_completed += 1
                                    _log_progress(
                                        retry_completed,
                                        len(batches),
                                        milestones,
                                        retry_shown,
                                        retry_start,
                                        logger,
                                    )
                            except Exception as e:
                                logger.error(
                                    f"Batch processing error (retry): {str(e)[:100]}"
                                )

                    # Log retry completion
                    retry_time = time.time() - retry_start
                    logger.info(
                        f"Retry complete: {retry_completed:,}/{len(batches):,} batches in {_format_time(retry_time)}"
                    )
                finally:
                    # Restore logger levels
                    fiona_logger.setLevel(old_fiona_level)
                    pyogrio_logger.setLevel(old_pyogrio_level)
            except Exception as validation_e:
                logger.error(
                    f"Failed to recover with validation: {str(validation_e)[:100]}"
                )
                return pd.DataFrame()

    if results:
        # Filter out empty DataFrames and all-NA columns to avoid FutureWarning in pd.concat
        results_filtered = []
        for df in results:
            if not df.empty:
                # Drop columns that are entirely NA
                df_clean = df.dropna(axis=1, how="all")
                if not df_clean.empty:
                    results_filtered.append(df_clean)
        results = results_filtered

        if results:
            # Concatenate with explicit dtype handling to suppress FutureWarning
            combined = pd.concat(results, ignore_index=True, sort=False)
            # Ensure all column names are strings (fixes pandas .str accessor issues later)
            combined.columns = combined.columns.astype(str)
        else:
            return pd.DataFrame()

        # Clean up duplicate external_id columns created by merges (if any exist)
        # external_id was already renamed during load, so we just need to handle duplicates
        if external_id_column and "external_id" in combined.columns:
            # Find merge duplicates like external_id_x, external_id_y, external_id_ee, external_id_client
            duplicate_variants = [
                col
                for col in combined.columns
                if col != "external_id" and col.startswith("external_id_")
            ]

            if duplicate_variants:
                logger.debug(
                    f"Dropping duplicate external_id columns: {duplicate_variants}"
                )
                combined = combined.drop(columns=duplicate_variants, errors="ignore")

        # plotId column is already present from batch processing
        # Just ensure it's at position 0
        if plot_id_column in combined.columns:
            combined = combined[
                [plot_id_column]
                + [col for col in combined.columns if col != plot_id_column]
            ]

        # Add admin context (Country, ProducerCountry, Admin_Level_1) from admin_code
        # MUST be done BEFORE formatting (which removes _median columns)
        logger.debug("Adding administrative context...")
        try:
            from openforis_whisp.parameters.lookup_gaul1_admin import lookup_dict

            combined = join_admin_codes(
                df=combined, lookup_dict=lookup_dict, id_col="admin_code_median"
            )
        except ImportError:
            logger.warning(
                "Could not import lookup dictionary - admin context not added"
            )

        # Format the output with error handling for bad bands
        logger.debug("Formatting output...")
        try:
            formatted = format_stats_dataframe(
                df=combined,
                area_col=f"{geometry_area_column}_sum",
                decimal_places=decimal_places,
                unit_type=unit_type,
                remove_columns=True,
                convert_water_flag=True,
            )
        except Exception as e:
            # If formatting fails, try recreating the image with validation
            logger.warning(
                f"Formatting failed: {str(e)[:100]}. Attempting to recreate image with band validation..."
            )
            try:
                with redirect_stdout(io.StringIO()):
                    whisp_image_validated = combine_datasets(
                        national_codes=national_codes, validate_bands=True
                    )

                # Reprocess batches with validated image - create a local process function
                logger.info("Reprocessing batches with validated image...")
                results_validated = []

                def process_batch_validated(
                    batch_idx: int, batch: gpd.GeoDataFrame
                ) -> Tuple[int, pd.DataFrame, pd.DataFrame]:
                    """Process one batch with validated image."""
                    with ee_semaphore:
                        fc = convert_batch_to_ee(batch)
                        if add_metadata_server:
                            fc = extract_centroid_and_geomtype_server(fc)
                        df_server = process_ee_batch(
                            fc,
                            whisp_image_validated,
                            reducer,
                            batch_idx,
                            max_retries,
                            logger,
                        )
                        df_client = extract_centroid_and_geomtype_client(
                            batch,
                            external_id_column=external_id_column,
                            return_attributes_only=True,
                        )
                    return batch_idx, df_server, df_client

                with ThreadPoolExecutor(max_workers=pool_workers) as executor:
                    futures = {
                        executor.submit(process_batch_validated, i, batch): i
                        for i, batch in enumerate(batches)
                    }

                    for future in as_completed(futures):
                        try:
                            batch_idx, df_server, df_client = future.result()
                            if plot_id_column not in df_server.columns:
                                logger.warning(
                                    f"Batch {batch_idx + 1} (retry): plotId DROPPED by EE. "
                                    f"Regenerating. Columns from EE: {list(df_server.columns)}"
                                )
                                # Use 1-indexed range to match client-side assignment
                                df_server[plot_id_column] = range(1, len(df_server) + 1)

                            # Ensure plotId is string type (consistent with creation)
                            if plot_id_column in df_server.columns:
                                df_server[plot_id_column] = df_server[
                                    plot_id_column
                                ].astype(str)
                            if plot_id_column in df_client.columns:
                                df_client[plot_id_column] = df_client[
                                    plot_id_column
                                ].astype(str)

                            # Drop external_id from df_server if it exists (already in df_client)
                            if "external_id" in df_server.columns:
                                df_server = df_server.drop(columns=["external_id"])

                            merged = df_server.merge(
                                df_client,
                                on=plot_id_column,
                                how="left",
                                suffixes=("", "_client"),
                            )
                            results_validated.append(merged)
                        except Exception as batch_e:
                            logger.error(
                                f"Batch reprocessing error: {str(batch_e)[:100]}"
                            )

                if results_validated:
                    # Concatenate with explicit dtype handling to suppress FutureWarning
                    combined = pd.concat(
                        results_validated, ignore_index=True, sort=False
                    )
                    # Ensure all column names are strings (fixes pandas .str accessor issues later)
                    combined.columns = combined.columns.astype(str)

                    # Clean up duplicate external_id columns created by merges (if any exist)
                    # external_id was already renamed during load, so we just need to handle duplicates
                    if external_id_column and "external_id" in combined.columns:
                        # Find merge duplicates like external_id_x, external_id_y, external_id_ee, external_id_client
                        duplicate_variants = [
                            col
                            for col in combined.columns
                            if col != "external_id" and col.startswith("external_id_")
                        ]

                        if duplicate_variants:
                            logger.debug(
                                f"Dropping duplicate external_id columns: {duplicate_variants}"
                            )
                            combined = combined.drop(
                                columns=duplicate_variants, errors="ignore"
                            )

                    # plotId column is already present, just ensure it's at position 0
                    if plot_id_column in combined.columns:
                        combined = combined[
                            [plot_id_column]
                            + [col for col in combined.columns if col != plot_id_column]
                        ]

                    # Add admin context again
                    try:
                        from openforis_whisp.parameters.lookup_gaul1_admin import (
                            lookup_dict,
                        )

                        combined = join_admin_codes(
                            df=combined,
                            lookup_dict=lookup_dict,
                            id_col="admin_code_median",
                        )
                    except ImportError:
                        logger.warning(
                            "Could not import lookup dictionary - admin context not added"
                        )

                    # Try formatting again with validated data
                    formatted = format_stats_dataframe(
                        df=combined,
                        area_col=f"{geometry_area_column}_sum",
                        decimal_places=decimal_places,
                        unit_type=unit_type,
                        remove_columns=True,
                        convert_water_flag=True,
                    )
                else:
                    logger.error(" Reprocessing with validation produced no results")
                    return pd.DataFrame()
            except Exception as retry_e:
                logger.error(
                    f"Failed to recover from formatting error: {str(retry_e)[:100]}"
                )
                raise retry_e

        # Ensure plot_id is present (should already be there from batch processing)
        if plot_id_column not in formatted.columns:
            logger.warning(f"{plot_id_column} column missing, regenerating...")
            formatted.insert(0, plot_id_column, range(1, len(formatted) + 1))

        # Note: Sorting is handled by format_stats_dataframe in the formatted wrapper functions

        logger.info(f"Processing complete: {len(formatted):,} features")
        return formatted
    else:
        logger.error(" No results produced")
        return pd.DataFrame()


# ============================================================================
# SEQUENTIAL PROCESSING (STANDARD ENDPOINT)
# ============================================================================


def whisp_stats_geojson_to_df_sequential(
    input_geojson_filepath: str,
    external_id_column: str = None,
    national_codes: List[str] = None,
    unit_type: str = "ha",
    whisp_image: ee.Image = None,
    custom_bands: Dict[str, Any] = None,
    add_metadata_client_side: bool = True,
    logger: logging.Logger = None,
    # Format parameters (auto-detect from config if not provided)
    decimal_places: int = None,
) -> pd.DataFrame:
    """
    Process GeoJSON sequentially using standard EE endpoint with automatic formatting.

    Uses reduceRegions for server-side processing and client-side metadata
    extraction via GeoPandas. Suitable for smaller datasets or when high-volume
    endpoint is not available. Automatically formats output.

    Requires: standard EE endpoint (default)

    Parameters
    ----------
    input_geojson_filepath : str
        Path to input GeoJSON
    external_id_column : str, optional
        Column name for external IDs
    national_codes : List[str], optional
        ISO2 codes for national datasets
    unit_type : str
        "ha" or "percent"
    whisp_image : ee.Image, optional
        Pre-combined image
    custom_bands : Dict[str, Any], optional
        Custom band information
    add_metadata_client_side : bool
        Add client-side metadata (recommended)
    logger : logging.Logger, optional
        Logger for output
    decimal_places : int, optional
        Decimal places for formatting. If None, auto-detects from config.

    Returns
    -------
    pd.DataFrame
        Formatted results DataFrame
    """
    from openforis_whisp.reformat import format_stats_dataframe

    logger = logger or logging.getLogger("whisp")

    # Suppress verbose output from dependencies (sequential has lower concurrency, use default)
    _suppress_verbose_output(max_concurrent=1)

    # Auto-detect decimal places from config if not provided
    if decimal_places is None:
        decimal_places = _extract_decimal_places(stats_area_columns_formatting)
        logger.debug(f"Using decimal_places={decimal_places} from config")

    # Validate endpoint
    validate_ee_endpoint("standard", raise_error=True)

    # Load GeoJSON with output suppressed (external_id_column renamed to 'external_id' if provided)
    gdf = _load_and_prepare_geojson(
        input_geojson_filepath, external_id_column=external_id_column
    )
    logger.info(f"Loaded {len(gdf):,} features")

    # Validate external_id if provided (lightweight client-side check)
    # Note: external_id_column already renamed to 'external_id' during load
    if external_id_column and "external_id" not in gdf.columns:
        # Exclude geometry column from available columns list
        available_cols = [c for c in gdf.columns if c != gdf.geometry.name]
        raise ValueError(
            f"Column '{external_id_column}' not found in GeoJSON properties. "
            f"Available columns: {available_cols}"
        )

    # Check completeness of external_id (warn if nulls exist)
    if external_id_column and "external_id" in gdf.columns:
        null_count = gdf["external_id"].isna().sum()
        if null_count > 0:
            null_pct = (null_count / len(gdf)) * 100
            logger.warning(
                f"Column 'external_id' (from '{external_id_column}') has {null_count:,} null values ({null_pct:.1f}% of {len(gdf):,} features). "
                f"These features may have missing external IDs in output."
            )

    # Clean geometries (preserve both null and invalid geometries by default)
    gdf = clean_geodataframe(
        gdf, remove_nulls=False, repair_geometries=False, logger=logger
    )

    # Add stable plotIds for merging (starting from 1, not 0)
    gdf[plot_id_column] = [str(i) for i in range(1, len(gdf) + 1)]

    # Strip unnecessary properties before sending to EE
    # Keep only: geometry, plot_id_column, and external_id
    # This prevents duplication of GeoJSON properties in EE results
    keep_cols = ["geometry", plot_id_column]
    if (
        external_id_column and "external_id" in gdf.columns
    ):  # Already renamed during load
        keep_cols.append("external_id")

    gdf_for_ee = gdf[keep_cols].copy()

    # CRITICAL: Convert external_id to string (both plotId and external_id are now strings)
    if external_id_column and "external_id" in gdf_for_ee.columns:
        gdf_for_ee["external_id"] = gdf_for_ee["external_id"].astype(str)
        logger.debug(f"Converted external_id column to string type")

    logger.debug(f"Stripped GeoJSON to essential columns: {keep_cols}")

    # Create image if not provided
    if whisp_image is None:
        logger.debug("Creating Whisp image...")
        # Suppress print statements from combine_datasets
        with redirect_stdout(io.StringIO()):
            try:
                # First try without validation
                whisp_image = combine_datasets(
                    national_codes=national_codes, validate_bands=False
                )
            except Exception as e:
                logger.warning(
                    f"First attempt failed: {str(e)[:100]}. Retrying with validate_bands=True..."
                )
                # Retry with validation to catch and fix bad bands
                whisp_image = combine_datasets(
                    national_codes=national_codes, validate_bands=True
                )

    # Drop external_id before sending to EE to enable caching
    # (external_id is preserved separately in gdf for client-side merging)
    gdf_for_ee_clean = gdf_for_ee.copy()
    if "external_id" in gdf_for_ee_clean.columns:
        gdf_for_ee_clean = gdf_for_ee_clean.drop(columns=["external_id"])
        logger.debug("Dropped external_id from data sent to EE (enables caching)")

    # Convert to EE (suppress print statements from convert_geojson_to_ee)
    logger.debug("Converting to EE FeatureCollection...")
    with redirect_stdout(io.StringIO()):
        fc = convert_geojson_to_ee(
            gdf_for_ee_clean, enforce_wgs84=True, strip_z_coords=True
        )

    # Create reducer
    reducer = ee.Reducer.sum().combine(ee.Reducer.median(), sharedInputs=True)

    # Process server-side with error handling for bad bands
    logger.info(
        f"Processing {len(gdf):,} features with Earth Engine (sequential mode)..."
    )
    try:
        results_fc = whisp_image.reduceRegions(collection=fc, reducer=reducer, scale=10)
        df_server = convert_ee_to_df(results_fc)
    except Exception as e:
        # Check if this is a band error
        error_msg = str(e)
        is_band_error = any(
            keyword in error_msg
            for keyword in ["Image.load", "asset", "not found", "does not exist"]
        )

        if is_band_error and whisp_image is not None:
            logger.warning(
                f"Detected bad band error: {error_msg[:100]}. Retrying with validate_bands=True..."
            )
            try:
                with redirect_stdout(io.StringIO()):
                    whisp_image = combine_datasets(
                        national_codes=national_codes, validate_bands=True
                    )
                logger.info("Image recreated with validation. Retrying processing...")
                results_fc = whisp_image.reduceRegions(
                    collection=fc, reducer=reducer, scale=10
                )
                df_server = convert_ee_to_df(results_fc)
            except Exception as retry_e:
                logger.error(f"Retry failed: {str(retry_e)[:100]}")
                raise
        else:
            raise

    logger.info("Server-side processing complete")

    # Ensure plotId is string type for consistent merges
    if plot_id_column in df_server.columns:
        df_server[plot_id_column] = df_server[plot_id_column].astype(str)

    # Add client-side metadata if requested
    if add_metadata_client_side:
        logger.debug("Extracting client-side metadata...")
        df_client = extract_centroid_and_geomtype_client(
            gdf,
            external_id_column=external_id_column,
            return_attributes_only=True,
        )

        # Ensure plotId is string type for consistent merges
        if plot_id_column in df_client.columns:
            df_client[plot_id_column] = df_client[plot_id_column].astype(str)

        # Drop external_id from df_server if it exists (keep from df_client - more reliable)
        if "external_id" in df_server.columns:
            df_server = df_server.drop(columns=["external_id"])

        # Merge on plotId (same strategy as concurrent mode)
        result = df_server.merge(
            df_client, on=plot_id_column, how="left", suffixes=("", "_client")
        )
    else:
        result = df_server

    # Format the output
    # Add admin context (Country, ProducerCountry, Admin_Level_1) from admin_code
    # MUST be done BEFORE formatting (which removes _median columns)
    logger.debug("Adding administrative context...")
    try:
        from openforis_whisp.parameters.lookup_gaul1_admin import lookup_dict

        result = join_admin_codes(
            df=result, lookup_dict=lookup_dict, id_col="admin_code_median"
        )
    except ImportError:
        logger.warning("Could not import lookup dictionary - admin context not added")

    # Format the output
    logger.debug("Formatting output...")
    formatted = format_stats_dataframe(
        df=result,
        area_col=f"{geometry_area_column}_sum",
        decimal_places=decimal_places,
        unit_type=unit_type,
        remove_columns=True,
        convert_water_flag=True,
    )

    # Ensure plot_id exists
    if plot_id_column not in formatted.columns:
        formatted.insert(0, plot_id_column, range(1, len(formatted) + 1))

    # Note: Sorting is handled by format_stats_dataframe in the formatted wrapper functions

    logger.info(f"Processing complete: {len(formatted):,} features")

    # external_id_column already renamed to 'external_id' during load - no action needed here

    return formatted


# ============================================================================
# FORMATTED WRAPPER FUNCTIONS (STATS + FORMAT)
# ============================================================================


def whisp_formatted_stats_geojson_to_df_concurrent(
    input_geojson_filepath: str,
    external_id_column: str = None,
    national_codes: List[str] = None,
    unit_type: str = "ha",
    whisp_image: ee.Image = None,
    custom_bands: Dict[str, Any] = None,
    batch_size: int = 10,
    max_concurrent: int = 20,
    validate_geometries: bool = True,
    max_retries: int = 3,
    add_metadata_server: bool = False,
    logger: logging.Logger = None,
    # Format parameters (auto-detect from config if not provided)
    decimal_places: int = None,
    remove_median_columns: bool = True,
    convert_water_flag: bool = True,
    water_flag_threshold: float = 0.5,
    sort_column: str = "plotId",
    geometry_audit_trail: bool = False,
) -> pd.DataFrame:
    """
    Process GeoJSON concurrently with automatic formatting and validation.

    Combines whisp_stats_geojson_to_df_concurrent + format_stats_dataframe + validation
    for a complete pipeline: extract stats → convert units → format output → validate schema.

    Uses high-volume endpoint and concurrent batching.

    Parameters
    ----------
    input_geojson_filepath : str
        Path to input GeoJSON file
    external_id_column : str, optional
        Column name for external IDs
    national_codes : List[str], optional
        ISO2 codes for national datasets
    unit_type : str
        "ha" or "percent"
    whisp_image : ee.Image, optional
        Pre-combined image
    custom_bands : Dict[str, Any], optional
        Custom band information
    batch_size : int
        Features per batch (default 25)
    max_concurrent : int
        Maximum concurrent EE calls (default 10)
    validate_geometries : bool
        Validate and clean geometries (default True)
    max_retries : int
        Retry attempts per batch (default 3)
    add_metadata_server : bool
        Add metadata server-side (default False)
    logger : logging.Logger, optional
        Logger for output
    decimal_places : int, optional
        Decimal places for rounding. If None, auto-detects from config:
        - Area columns: geometry_area_column_formatting
        - Percent columns: stats_percent_columns_formatting
        - Other columns: stats_area_columns_formatting
    remove_median_columns : bool
        Remove '_median' columns (default True)
    convert_water_flag : bool
        Convert water flag to boolean (default True)
    water_flag_threshold : float
        Water flag ratio threshold (default 0.5)
    sort_column : str
        Column to sort by (default "plotId", None to skip)
    geometry_audit_trail : bool, default False
        If True, includes original input geometry column:
        - geo_original: Original input geometry (before EE processing), stored as GeoJSON
        Enables geometry traceability for compliance and audit purposes.

    Returns
    -------
    pd.DataFrame
        Validated, formatted results DataFrame with optional audit trail
    """
    from openforis_whisp.reformat import format_stats_dataframe
    from datetime import datetime, timezone
    import json
    from shapely.geometry import mapping

    logger = logger or logging.getLogger("whisp")

    # Auto-detect decimal places from config if not provided
    if decimal_places is None:
        # Use stats_area_columns_formatting as default for most columns
        decimal_places = _extract_decimal_places(stats_area_columns_formatting)
        logger.debug(f"Using decimal_places={decimal_places} from config")

    # Load original geometries once here if needed for audit trail (avoid reloading later)
    gdf_original_geoms = None
    if geometry_audit_trail:
        logger.debug("Pre-loading GeoJSON for geometry audit trail...")
        gdf_original_geoms = _load_and_prepare_geojson(input_geojson_filepath)

    # Step 1: Get raw stats
    logger.debug("Step 1/2: Extracting statistics (concurrent)...")
    df_raw = whisp_stats_geojson_to_df_concurrent(
        input_geojson_filepath=input_geojson_filepath,
        external_id_column=external_id_column,
        national_codes=national_codes,
        unit_type=unit_type,
        whisp_image=whisp_image,
        custom_bands=custom_bands,
        batch_size=batch_size,
        max_concurrent=max_concurrent,
        validate_geometries=validate_geometries,
        max_retries=max_retries,
        add_metadata_server=add_metadata_server,
        logger=logger,
    )

    # Step 2: Format the output
    logger.debug("Step 2/2: Formatting output...")
    median_cols_before = [c for c in df_raw.columns if c.endswith("_median")]
    logger.debug(
        f"Columns ending with '_median' BEFORE formatting: {median_cols_before}"
    )

    df_formatted = format_stats_dataframe(
        df=df_raw,
        area_col=f"{geometry_area_column}_sum",
        decimal_places=decimal_places,
        unit_type=unit_type,
        remove_columns=remove_median_columns,
        convert_water_flag=convert_water_flag,
        water_flag_threshold=water_flag_threshold,
        sort_column=sort_column,
    )

    median_cols_after = [c for c in df_formatted.columns if c.endswith("_median")]
    logger.debug(f"Columns ending with '_median' AFTER formatting: {median_cols_after}")

    # Step 2b: Reformat geometry and handle point areas
    try:
        df_formatted = reformat_geometry_type(df_formatted)
    except Exception as e:
        logger.warning(f"Error reformatting geometry type: {e}")

    try:
        df_formatted = set_point_geometry_area_to_zero(df_formatted)
    except Exception as e:
        logger.warning(f"Error setting point geometry area to zero: {e}")

    # Step 3: Validate against schema
    logger.debug("Step 3/3: Validating against schema...")
    from openforis_whisp.reformat import validate_dataframe_using_lookups_flexible

    df_validated = validate_dataframe_using_lookups_flexible(
        df_stats=df_formatted,
        national_codes=national_codes,
        custom_bands=custom_bands,
    )

    # Step 2c: Add audit trail column (AFTER validation to preserve columns)
    if geometry_audit_trail:
        logger.debug("Adding geo_original column for audit trail...")
        df_validated = _add_geometry_audit_trail(
            df_validated=df_validated,
            input_geojson_filepath=input_geojson_filepath,
            gdf_original_geoms=gdf_original_geoms,
            logger=logger,
        )

    # Add processing metadata column using pd.concat to avoid fragmentation warning
    metadata_dict = {
        "whisp_version": get_version("openforis-whisp"),
        "processing_timestamp_utc": datetime.now(timezone.utc).strftime(
            "%Y-%m-%d %H:%M:%S%z"
        ),
    }
    metadata_series = pd.Series(
        [metadata_dict] * len(df_validated), name="whisp_processing_metadata"
    )
    df_validated = pd.concat([df_validated, metadata_series], axis=1)

    logger.info("Concurrent processing + formatting + validation complete")
    return df_validated


def whisp_formatted_stats_geojson_to_df_sequential(
    input_geojson_filepath: str,
    external_id_column: str = None,
    national_codes: List[str] = None,
    unit_type: str = "ha",
    whisp_image: ee.Image = None,
    custom_bands: Dict[str, Any] = None,
    add_metadata_client_side: bool = True,
    logger: logging.Logger = None,
    # Format parameters (auto-detect from config if not provided)
    decimal_places: int = None,
    remove_median_columns: bool = True,
    convert_water_flag: bool = True,
    water_flag_threshold: float = 0.5,
    sort_column: str = "plotId",
    geometry_audit_trail: bool = False,
) -> pd.DataFrame:
    """
    Process GeoJSON sequentially with automatic formatting and validation.

    Combines whisp_stats_geojson_to_df_sequential + format_stats_dataframe + validation
    for a complete pipeline: extract stats → convert units → format output → validate schema.

    Uses standard endpoint for sequential processing.

    Parameters
    ----------
    input_geojson_filepath : str
        Path to input GeoJSON file
    external_id_column : str, optional
        Column name for external IDs
    national_codes : List[str], optional
        ISO2 codes for national datasets
    unit_type : str
        "ha" or "percent"
    whisp_image : ee.Image, optional
        Pre-combined image
    custom_bands : Dict[str, Any], optional
        Custom band information
    add_metadata_client_side : bool
        Add client-side metadata (default True)
    logger : logging.Logger, optional
        Logger for output
    decimal_places : int, optional
        Decimal places for rounding. If None, auto-detects from config:
        - Area columns: geometry_area_column_formatting
        - Percent columns: stats_percent_columns_formatting
        - Other columns: stats_area_columns_formatting
    remove_median_columns : bool
        Remove '_median' columns (default True)
    convert_water_flag : bool
        Convert water flag to boolean (default True)
    water_flag_threshold : float
        Water flag ratio threshold (default 0.5)
    sort_column : str
        Column to sort by (default "plotId", None to skip)
    geometry_audit_trail : bool, default True
        If True, includes original input geometry column:
        - geo_original: Original input geometry (before EE processing), stored as GeoJSON
        Enables geometry traceability for compliance and audit purposes.

    Returns
    -------
    pd.DataFrame
        Validated, formatted results DataFrame with optional audit trail
    """
    from openforis_whisp.reformat import format_stats_dataframe
    from datetime import datetime, timezone
    import json
    from shapely.geometry import mapping

    logger = logger or logging.getLogger("whisp")

    # Auto-detect decimal places from config if not provided
    if decimal_places is None:
        # Use stats_area_columns_formatting as default for most columns
        decimal_places = _extract_decimal_places(stats_area_columns_formatting)
        logger.debug(f"Using decimal_places={decimal_places} from config")

    # Load original geometries once here if needed for audit trail (avoid reloading later)
    gdf_original_geoms = None
    if geometry_audit_trail:
        logger.debug("Pre-loading GeoJSON for geometry audit trail...")
        gdf_original_geoms = _load_and_prepare_geojson(input_geojson_filepath)

    # Step 1: Get raw stats
    logger.debug("Step 1/2: Extracting statistics (sequential)...")
    df_raw = whisp_stats_geojson_to_df_sequential(
        input_geojson_filepath=input_geojson_filepath,
        external_id_column=external_id_column,
        national_codes=national_codes,
        unit_type=unit_type,
        whisp_image=whisp_image,
        custom_bands=custom_bands,
        add_metadata_client_side=add_metadata_client_side,
        logger=logger,
    )

    # Step 2: Format the output
    logger.debug("Step 2/2: Formatting output...")
    median_cols_before = [c for c in df_raw.columns if c.endswith("_median")]
    logger.debug(
        f"Columns ending with '_median' BEFORE formatting: {median_cols_before}"
    )

    df_formatted = format_stats_dataframe(
        df=df_raw,
        area_col=f"{geometry_area_column}_sum",
        decimal_places=decimal_places,
        unit_type=unit_type,
        remove_columns=remove_median_columns,
        convert_water_flag=convert_water_flag,
        water_flag_threshold=water_flag_threshold,
        sort_column=sort_column,
    )

    median_cols_after = [c for c in df_formatted.columns if c.endswith("_median")]
    logger.debug(f"Columns ending with '_median' AFTER formatting: {median_cols_after}")

    # Step 2b: Reformat geometry and handle point areas
    try:
        df_formatted = reformat_geometry_type(df_formatted)
    except Exception as e:
        logger.warning(f"Error reformatting geometry type: {e}")

    try:
        df_formatted = set_point_geometry_area_to_zero(df_formatted)
    except Exception as e:
        logger.warning(f"Error setting point geometry area to zero: {e}")

    # Step 3: Validate against schema
    logger.debug("Step 3/3: Validating against schema...")
    from openforis_whisp.reformat import validate_dataframe_using_lookups_flexible

    df_validated = validate_dataframe_using_lookups_flexible(
        df_stats=df_formatted,
        national_codes=national_codes,
        custom_bands=custom_bands,
    )

    # Step 2c: Add audit trail column (AFTER validation to preserve columns)
    if geometry_audit_trail:
        logger.debug("Adding geo_original column for audit trail...")
        df_validated = _add_geometry_audit_trail(
            df_validated=df_validated,
            input_geojson_filepath=input_geojson_filepath,
            gdf_original_geoms=gdf_original_geoms,
            logger=logger,
        )

    # Add processing metadata column using pd.concat to avoid fragmentation warning
    metadata_dict = {
        "whisp_version": get_version("openforis-whisp"),
        "processing_timestamp_utc": datetime.now(timezone.utc).strftime(
            "%Y-%m-%d %H:%M:%S%z"
        ),
    }
    metadata_series = pd.Series(
        [metadata_dict] * len(df_validated), name="whisp_processing_metadata"
    )
    df_validated = pd.concat([df_validated, metadata_series], axis=1)

    logger.info("Sequential processing + formatting + validation complete")
    return df_validated


# ============================================================================
# FAST PROCESSING WITH AUTO-ROUTING
# ============================================================================


def whisp_formatted_stats_geojson_to_df_fast(
    input_geojson_filepath: str,
    external_id_column: str = None,
    national_codes: List[str] = None,
    unit_type: str = "ha",
    whisp_image: ee.Image = None,
    custom_bands: Dict[str, Any] = None,
    mode: str = "sequential",
    # Concurrent-specific parameters
    batch_size: int = 10,
    max_concurrent: int = 20,
    validate_geometries: bool = True,
    max_retries: int = 3,
    add_metadata_server: bool = False,
    # Format parameters (auto-detect from config if not provided)
    decimal_places: int = None,
    remove_median_columns: bool = True,
    convert_water_flag: bool = True,
    water_flag_threshold: float = 0.5,
    sort_column: str = "plotId",
    geometry_audit_trail: bool = False,
) -> pd.DataFrame:
    """
    Process GeoJSON to Whisp statistics with optimized fast processing.

    Routes to concurrent (high-volume endpoint) or sequential (standard endpoint)
    based on explicit mode selection.

    This is the recommended entry point for most users.

    Parameters
    ----------
    input_geojson_filepath : str
        Path to input GeoJSON file
    external_id_column : str, optional
        Column name for external IDs
    national_codes : List[str], optional
        ISO2 codes for national datasets
    unit_type : str
        "ha" or "percent"
    whisp_image : ee.Image, optional
        Pre-combined image
    custom_bands : Dict[str, Any], optional
        Custom band information
    mode : str
        Processing mode:
        - "concurrent": Uses high-volume endpoint with batch processing
        - "sequential": Uses standard endpoint for sequential processing
    batch_size : int
        Features per batch (only for concurrent mode)
    max_concurrent : int
        Maximum concurrent EE calls (only for concurrent mode)
    validate_geometries : bool
        Validate and clean geometries
    max_retries : int
        Retry attempts per batch (only for concurrent mode)
    add_metadata_server : bool
        Add metadata server-side (only for concurrent mode)
    decimal_places : int, optional
        Decimal places for rounding. If None, auto-detects from config.
    remove_median_columns : bool
        Remove '_median' columns
    convert_water_flag : bool
        Convert water flag to boolean
    water_flag_threshold : float
        Water flag ratio threshold
    sort_column : str
        Column to sort by
    geometry_audit_trail : bool
        Include geometry modification audit trail columns

    Returns
    -------
    pd.DataFrame
        Validated, formatted results DataFrame

    Examples
    --------
    >>> # Use concurrent processing (recommended for most datasets)
    >>> df = whisp_formatted_stats_geojson_to_df_fast(
    ...     "data.geojson",
    ...     mode="concurrent"
    ... )

    >>> # Use sequential processing for more stable results
    >>> df = whisp_formatted_stats_geojson_to_df_fast(
    ...     "data.geojson",
    ...     mode="sequential"
    ... )
    """
    logger = logging.getLogger("whisp")

    # Validate mode parameter
    if mode not in ("concurrent", "sequential"):
        raise ValueError(
            f"Invalid mode '{mode}'. Must be 'concurrent' or 'sequential'."
        )

    logger.info(f"Mode: {mode}")

    # Route to appropriate function
    if mode == "concurrent":
        logger.debug("Routing to concurrent processing...")
        return whisp_formatted_stats_geojson_to_df_concurrent(
            input_geojson_filepath=input_geojson_filepath,
            external_id_column=external_id_column,
            national_codes=national_codes,
            unit_type=unit_type,
            whisp_image=whisp_image,
            custom_bands=custom_bands,
            batch_size=batch_size,
            max_concurrent=max_concurrent,
            validate_geometries=validate_geometries,
            max_retries=max_retries,
            add_metadata_server=add_metadata_server,
            logger=logger,
            decimal_places=decimal_places,
            remove_median_columns=remove_median_columns,
            convert_water_flag=convert_water_flag,
            water_flag_threshold=water_flag_threshold,
            sort_column=sort_column,
            geometry_audit_trail=geometry_audit_trail,
        )
    else:  # sequential
        logger.debug("Routing to sequential processing...")
        return whisp_formatted_stats_geojson_to_df_sequential(
            input_geojson_filepath=input_geojson_filepath,
            external_id_column=external_id_column,
            national_codes=national_codes,
            unit_type=unit_type,
            whisp_image=whisp_image,
            custom_bands=custom_bands,
            logger=logger,
            decimal_places=decimal_places,
            remove_median_columns=remove_median_columns,
            convert_water_flag=convert_water_flag,
            water_flag_threshold=water_flag_threshold,
            sort_column=sort_column,
            geometry_audit_trail=geometry_audit_trail,
        )
